!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_input_mod
!> @brief   read parameters and data for md simulation
!! @authors Yuji Sugita (YS), Takaharu Mori (TM), Chigusa Kobayashi (CK), 
!!          Jaewoon Jung (JJ), Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_input_mod

  use at_remd_str_mod
  use fileio_control_mod
  use fileio_rst_mod
  use fileio_psf_mod
  use fileio_gpr_mod
  use fileio_str_mod
  use fileio_par_mod
  use fileio_top_mod
  use fileio_crd_mod
  use fileio_pdb_mod
  use fileio_control_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod

  implicit none
  private

  ! structures
  type, public :: s_inp_info
    character(MaxFilename) :: topfile    = ''
    character(MaxFilename) :: parfile    = ''
    character(MaxFilename) :: strfile    = ''
    character(MaxFilename) :: gprfile    = ''
    character(MaxFilename) :: psffile    = ''
    character(MaxFilename) :: pdbfile    = ''
    character(MaxFilename) :: crdfile    = ''
    character(MaxFilename) :: rstfile    = ''
    character(MaxFilename) :: reffile    = ''
  end type s_inp_info

  ! subroutines
  public  :: show_ctrl_input
  public  :: read_ctrl_input
  public  :: input_md
  public  :: input_min
  public  :: input_remd
  private :: include_id_to_filename

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_input
  !> @brief        show INPUT section usage
  !! @authors      NT
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_input(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('md')

        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') '##  CHARMM Force Field'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'strfile = sample.str      # stream file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample.pdb      # PDB file'
        write(MsgOut,'(A)') '# crdfile = sample.crd      # coordinates file'
        write(MsgOut,'(A)') '# rstfile = sample.rst      # restart file'
        write(MsgOut,'(A)') '# reffile = sample.pdb      # reference for restraint'
        write(MsgOut,'(A)') ' '


      case ('min')

        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') '##  CHARMM Force Field'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'strfile = sample.str      # stream file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample.pdb      # PDB file'
        write(MsgOut,'(A)') '# crdfile = sample.crd      # coordinates file'
        write(MsgOut,'(A)') '# rstfile = sample.rst      # restart file'
        write(MsgOut,'(A)') '# reffile = sample.pdb      # reference for restraint'
        write(MsgOut,'(A)') ' '


      case ('remd')
        
        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') '## Force Field'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'strfile = sample.str      # stream file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample{}.pdb    # PDB file'
        write(MsgOut,'(A)') '# crdfile = sample{}.crd    # coordinates file'
        write(MsgOut,'(A)') '# rstfile = sample{}.rst    # restart file'
        write(MsgOut,'(A)') '# reffile = sample.pdb      # reference for restraint'
        write(MsgOut,'(A)') ' '


      end select

    else

      select case (run_mode)

      case ('md')

        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample.pdb      # PDB file'
        write(MsgOut,'(A)') ' '


      case ('min')

        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample.pdb      # PDB file'
        write(MsgOut,'(A)') ' '


      case ('remd')

        write(MsgOut,'(A)') '[INPUT]'
        write(MsgOut,'(A)') 'topfile = sample.top      # topology file'
        write(MsgOut,'(A)') 'parfile = sample.par      # parameter file'
        write(MsgOut,'(A)') 'psffile = sample.psf      # protein structure file'
        write(MsgOut,'(A)') 'pdbfile = sample{}.pdb    # PDB file'
        write(MsgOut,'(A)') ' '


      end select

    end if

    return

  end subroutine show_ctrl_input
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_input
  !> @brief        read INPUT section in the control file
  !! @authors      YS, TM, JJ, CK
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   inp_info : INPUT section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_input(handle, inp_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Input'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_inp_info),        intent(inout) :: inp_info

    
    ! read parameters from control file
    ! 
    call read_ctrlfile_string(handle, Section, 'Topfile',   inp_info%topfile)
    call read_ctrlfile_string(handle, Section, 'Parfile',   inp_info%parfile)
    call read_ctrlfile_string(handle, Section, 'Strfile',   inp_info%strfile)
    call read_ctrlfile_string(handle, Section, 'Gprfile',   inp_info%gprfile)
    call read_ctrlfile_string(handle, Section, 'Psffile',   inp_info%psffile)
    call read_ctrlfile_string(handle, Section, 'Pdbfile',   inp_info%pdbfile)
    call read_ctrlfile_string(handle, Section, 'Crdfile',   inp_info%crdfile)
    call read_ctrlfile_string(handle, Section, 'Rstfile',   inp_info%rstfile)
    call read_ctrlfile_string(handle, Section, 'Reffile',   inp_info%reffile)

    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Ctrl_Input> Input Files'

      if (inp_info%topfile /= '') then
        write(MsgOut,*) ' topfile = ', trim(inp_info%topfile)
      end if

      if (inp_info%parfile /= '') then
        write(MsgOut,*) ' parfile = ', trim(inp_info%parfile)
      end if

      if (inp_info%strfile /= '') then
        write(MsgOut,*) ' strfile = ', trim(inp_info%strfile)
      end if

      if (inp_info%gprfile /= '') then
        write(MsgOut,*) ' gprfile = ', trim(inp_info%gprfile)
      end if

      if (inp_info%psffile /= '') then
        write(MsgOut,*) ' psffile = ', trim(inp_info%psffile)
      end if

      if (inp_info%crdfile /= '') then
        write(MsgOut,*) ' crdfile = ', trim(inp_info%crdfile)
      end if

      if (inp_info%pdbfile /= '') then
        write(MsgOut,*) ' pdbfile = ', trim(inp_info%pdbfile)
      end if

      if (inp_info%rstfile /= '') then
        write(MsgOut,*) ' rstfile = ', trim(inp_info%rstfile)
      end if

      if (inp_info%reffile /= '') then
        write(MsgOut,*) ' reffile = ', trim(inp_info%reffile)
      end if

      write(MsgOut,'(A)') ' '

    end if

    return

  end subroutine read_ctrl_input

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    input_md
  !> @brief        read input data from files
  !! @authors      YS, TM, CK, NT
  !! @param[in]    inp_info : information of INPUT section control parameters
  !! @param[out]   top      : information of CHARMM topology data
  !! @param[out]   par      : information of CHARMM force field parameters
  !! @param[out]   gpr      : information of GENESIS Go model parameters
  !! @param[out]   psf      : information of Protein structure data
  !! @param[out]   pdb      : information of coordinate data
  !! @param[out]   crd      : information of coordinate data
  !! @param[out]   rst      : information of restart data
  !! @param[out]   ref      : information of reference coordinate data
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_md(inp_info, top, par, gpr, psf, pdb, crd, rst, ref)

    ! formal arguments
    type(s_inp_info),        intent(in)    :: inp_info
    type(s_top),             intent(inout) :: top
    type(s_par),             intent(inout) :: par
    type(s_gpr),             intent(inout) :: gpr
    type(s_psf),             intent(inout) :: psf
    type(s_pdb),             intent(inout) :: pdb
    type(s_crd),             intent(inout) :: crd
    type(s_rst),             intent(inout) :: rst
    type(s_pdb),             intent(inout) :: ref


    if (inp_info%topfile /= '') then
      call input_top(inp_info%topfile, top)
    end if

    if (inp_info%parfile /= '') then
      call input_par(inp_info%parfile, par, top)
    end if

    if (inp_info%strfile /= '') then
      call input_str(inp_info%strfile, top, par)
    end if

    if (inp_info%gprfile /= '') then
      call input_gpr(inp_info%gprfile, gpr)
    end if

    if (inp_info%psffile /= '') then
      call input_psf(inp_info%psffile, psf)
    end if

    if (inp_info%crdfile /= '') then
      call input_crd(inp_info%crdfile, crd)
    end if

    if (inp_info%pdbfile /= '') then
      call input_pdb(inp_info%pdbfile, pdb)
    end if

    rst%rstfile_type = RstfileTypeUndef
    if (inp_info%rstfile /= '') then
      call input_rst(inp_info%rstfile, rst)
    end if

    if (inp_info%reffile /= '') then
      call input_pdb(inp_info%reffile, ref)
    end if

    return

  end subroutine input_md

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    input_min
  !> @brief        read input data from files
  !! @authors      TM, CK
  !! @param[in]    inp_info : information of INPUT section control parameters
  !! @param[out]   top      : information of CHARMM topology data
  !! @param[out]   par      : information of CHARMM force field parameters
  !! @param[out]   gpr      : information of GENESIS Go model parameters
  !! @param[out]   psf      : information of Protein structure data
  !! @param[out]   pdb      : information of coordinate data
  !! @param[out]   crd      : information of coordinate data
  !! @param[out]   rst      : information of restart data
  !! @param[out]   ref      : information of reference coordinate data
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_min(inp_info, top, par, gpr, psf, pdb, crd, rst, ref)

    ! formal arguments
    type(s_inp_info),        intent(in)    :: inp_info
    type(s_top),             intent(inout) :: top
    type(s_par),             intent(inout) :: par
    type(s_gpr),             intent(inout) :: gpr
    type(s_psf),             intent(inout) :: psf
    type(s_pdb),             intent(inout) :: pdb
    type(s_crd),             intent(inout) :: crd
    type(s_rst),             intent(inout) :: rst
    type(s_pdb),             intent(inout) :: ref


    if (inp_info%topfile /= '') then
      call input_top(inp_info%topfile, top)
    end if

    if (inp_info%parfile /= '') then
      call input_par(inp_info%parfile, par, top)
    end if

    if (inp_info%strfile /= '') then
      call input_str(inp_info%strfile, top, par)
    end if

    if (inp_info%gprfile /= '') then
      call input_gpr(inp_info%gprfile, gpr)
    end if

    if (inp_info%psffile /= '') then
      call input_psf(inp_info%psffile, psf)
    end if

    if (inp_info%crdfile /= '') then
      call input_crd(inp_info%crdfile, crd)
    end if

    if (inp_info%pdbfile /= '') then
      call input_pdb(inp_info%pdbfile, pdb)
    end if

    rst%rstfile_type = RstfileTypeUndef
    if (inp_info%rstfile /= '') then
      call input_rst(inp_info%rstfile, rst)
    end if

    if (inp_info%reffile /= '') then
      call input_pdb(inp_info%reffile, ref)
    end if

    return

  end subroutine input_min

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    input_remd
  !> @brief        read input data from files
  !! @authors      TM, CK
  !! @param[in]    inp_info : information of INPUT section control parameters
  !! @param[in]    remd     : information of REMD
  !! @param[out]   top      : information of CHARMM topology data
  !! @param[out]   par      : information of CHARMM force field parameters
  !! @param[out]   gpr      : information of GENESIS Go model parameters
  !! @param[out]   psf      : information of Protein structure data
  !! @param[out]   pdb      : information of coordinate data
  !! @param[out]   crd      : information of coordinate data
  !! @param[out]   rst      : information of restart data
  !! @param[out]   ref      : information of reference coordinate data
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_remd(inp_info, top, par, gpr, psf, pdb, crd, rst, ref)

    ! formal arguments
    type(s_inp_info),         intent(in)    :: inp_info
    type(s_top),              intent(inout) :: top
    type(s_par),              intent(inout) :: par
    type(s_gpr),              intent(inout) :: gpr
    type(s_psf),              intent(inout) :: psf
    type(s_pdb),              intent(inout) :: pdb
    type(s_crd),              intent(inout) :: crd
    type(s_rst),              intent(inout) :: rst
    type(s_pdb),              intent(inout) :: ref

    ! local variables
    character(MaxMultiFilename)             :: filename


    if (inp_info%topfile /= '') then
      filename = inp_info%topfile
      call input_top(filename, top)
    end if

    if (inp_info%parfile /= '') then
      filename = inp_info%parfile
      call input_par(filename, par, top)
    end if

    if (inp_info%strfile /= '') then
      filename = inp_info%strfile
      call input_str(filename, top, par)
    end if

    if (inp_info%gprfile /= '') then
      filename = inp_info%gprfile
      call input_gpr(filename, gpr)
    end if

    if (inp_info%psffile /= '') then
      filename = inp_info%psffile
      call input_psf(filename, psf)
    end if

    if (inp_info%reffile /= '') then
      filename = inp_info%reffile
      call include_id_to_filename(filename)
      call input_pdb(filename, ref)
    end if

    if (inp_info%crdfile /= '') then
      filename = inp_info%crdfile
      call include_id_to_filename(filename)
      call input_crd(filename, crd)
    end if

    if (inp_info%pdbfile /= '') then
      filename = inp_info%pdbfile
      call include_id_to_filename(filename)
      call input_pdb(filename, pdb)
    end if

    rst%rstfile_type = RstfileTypeUndef
    if (inp_info%rstfile /= '') then
      filename = inp_info%rstfile
      call include_id_to_filename(filename)
      call input_rst(filename, rst)
    end if

    return

  end subroutine input_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    include_id_to_filename
  !> @brief        include id to filename
  !! @authors      TM
  !! @param[inout] filename : replicate filename
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine include_id_to_filename(filename)

    ! formal arguments
    character(MaxFilename),  intent(inout) :: filename

    ! local variables
    integer                  :: comp, ndigit, id
    integer                  :: i, j, ci1, ci2, cnumber
    character(MaxFilename)   :: filename_ori
    character(10)            :: frmt, cd, cid


    ! define replica id
    !
    id = my_country_no + 1
    do i = 1, 100
      comp = 10**i
      if (id < comp) then
        ndigit = i
        exit
      end if
    end do

    ! check filename
    !
    filename_ori = filename

    ci1 = scan(filename, '{')
    ci2 = scan(filename, '}')

    if (ci1 == 0 .or. ci2 ==0) &
      return

    if (ci1 > 0 .and. ci2 > ci1) then

      write(cd,'(i10)') ndigit
      frmt = '(i' // trim(adjustl(cd)) // '.' // trim(adjustl(cd)) // ')'
      write(cid,frmt) id

      cnumber = len_trim(filename_ori)
      if (cnumber + ndigit > MaxFilename) &
         call error_msg('Error: too long filename'//filename_ori)

      j = 0
      do i = 1, ci1 - 1
        j = j + 1
        filename(j:j) = filename_ori(i:i)
      end do
      do i = 1, ndigit
        j = j + 1
        filename(j:j) = cid(i:i)
      end do
      do i = ci2+1, MaxFilename
        j = j + 1
        filename(j:j) = filename_ori(i:i)
      end do

    end if

    return

  end subroutine include_id_to_filename

end module at_input_mod
