!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   constants_mod
!> @brief   constant variables (this module can be used like common in F77)
!! @authors Yuji Sugita (YS)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module constants_mod

#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! constants for real precision
  integer,  public, parameter :: sp = selected_real_kind(6, 37)
  integer,  public, parameter :: dp = selected_real_kind(15, 307)

#ifdef _SINGLE

  integer,  public, parameter :: wp = sp

#ifdef MPI

  integer,  public, parameter :: mpi_wp_real    = MPI_Real4
  integer,  public, parameter :: mpi_wp_complex = MPI_Complex

#endif

#else

  integer,  public, parameter :: wp = dp

#ifdef MPI

  integer,  public, parameter :: mpi_wp_real    = MPI_Real8
  integer,  public, parameter :: mpi_wp_complex = MPI_Double_complex

#endif

#endif

  ! constants
  real(wp), public, parameter :: PI         = 3.14159265358979323846_wp
  real(wp), public, parameter :: RAD        = PI/180.0_wp
  real(wp), public, parameter :: EPS        = 1.0e-10_wp
  real(wp), public, parameter :: HALF       = 0.5_wp
  real(wp), public, parameter :: ONE_THIRD  = 1.0_wp/3.0_wp
  real(wp), public, parameter :: AKMA_PS    = 4.88882129e-02_wp
  real(wp), public, parameter :: ELECOEF    = 332.0716_wp
  real(wp), public, parameter :: J_CAL      = 4.18605_wp
  real(wp), public, parameter :: KBOLTZ     = 1.987191e-03_wp
  real(wp), public, parameter :: ATMOS_P    = 1.4584007e-05_wp
  real(wp), public, parameter :: P_ATMOS    = 1.0_wp/ATMOS_P

end module constants_mod
