!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   pr_setup_spdyn_peer_mod
!> @brief   peer module of lib/domain_setup_md_mod
!! @authors Norio Takase (NT), Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module pr_setup_spdyn_peer_mod

  use pr_select_atoms_peer_mod
  use pr_control_mod
  use pr_domain_index_file_mod
  use pr_cached_molecule_mod
  use sp_dynamics_mod
  use sp_dynvars_mod
  use sp_domain_mod
  use sp_restraints_mod
  use sp_constraints_mod
  use sp_enefunc_restraints_mod
  use sp_energy_mod
  use sp_boundary_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_restraints_str_mod
  use sp_constraints_str_mod
  use sp_boundary_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use dihedral_libs_mod
  use select_mod
  use select_atoms_str_mod
  use fileio_par_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
  use omp_lib

  implicit none
  private

  ! subroutines
  public  :: setup_md_peer_0
  private ::   setup_boundary_peer_0
  private ::   setup_domain_peer_0
  private ::     setup_solute_and_water_peer_0
  private ::     setup_hbond_group_peer_0
  private ::   setup_restraints_peer_0
  private ::   define_enefunc_peer_0
  private ::     setup_enefunc_nonb_peer_0
  private ::     setup_enefunc_table_peer_0
  private ::     setup_enefunc_restraints_peer_0
  private ::       setup_enefunc_rest_group_peer_0
  private ::       setup_enefunc_rest_func_peer_0
  private ::   setup_dynamics_peer_0
  private ::   setup_constraints_peer_0
  private ::     setup_fast_water_peer_0
  private ::     setup_rigid_bond_peer_0

  public  :: setup_md_peer
  private ::   setup_domain_peer
  private ::     setup_atom_by_HBond_peer
  private ::     setup_atom_by_table_peer
  private ::     setup_atom_peer
  private ::       molecule_to_domain_peer
  private ::   define_enefunc_peer
  private ::     setup_enefunc_bond_peer
  private ::     setup_enefunc_angl_peer
  private ::     setup_enefunc_dihe_peer
  private ::     setup_enefunc_impr_peer
  private ::     setup_enefunc_cmap_peer
  private ::     setup_enefunc_bond_constraint_peer
  private ::     setup_enefunc_restraints_peer
  private ::       setup_enefunc_rest_domain_peer
  private ::     alloc_id_g2l
  
contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_md_peer_0
  !> @brief        peer routine of sp_setup_spdyn_mod::setup_md()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_md_peer_0(ctrl_data,    &
                             par,          &
                             boundary_rst, &
                             boundary,     &
                             enefunc,      &
                             dynvars,      &
                             dynamics,     &
                             constraints,  &
                             restraints)

    ! formal arguments
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_par),             intent(in)    :: par
    type(s_boundary),        intent(in)    :: boundary_rst
    type(s_boundary),        intent(inout) :: boundary
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_dynvars),         intent(inout) :: dynvars
    type(s_dynamics),        intent(inout) :: dynamics
    type(s_constraints),     intent(inout) :: constraints
    type(s_restraints),      intent(inout) :: restraints


#ifdef OMP
    if (omp_get_thread_num() /= 0) &
      call error_msg('Setup_Md_Peer_0> '//&
                     'multi-thread is not supported.')
#endif

    ! set parameters for boundary condition
    !
    call setup_boundary_peer_0(ctrl_data%bou_info,              &
                               ctrl_data%ene_info%table,        &
                               ctrl_data%ene_info%pairlistdist, &
                               ctrl_data%ene_info%water_model,  &
                               ctrl_data%ens_info%ensemble,     &
                               ctrl_data%cons_info%rigid_bond,  &
                               boundary_rst,                    &
                               boundary)

    ! set parameters for domain 
    !
    call setup_domain_peer_0(ctrl_data%ene_info,  &
                             ctrl_data%cons_info, &
                             par, boundary, enefunc, constraints)


    ! set parameters for communication
    !
    ! ..skip


    ! set parameters for restraints
    !
    call setup_restraints_peer_0(ctrl_data%res_info, &
                                 ctrl_data%sel_info, &
                                 restraints)


    ! setup enefunc in each domain
    !
    call define_enefunc_peer_0(ctrl_data%ene_info, &
                               par, restraints, enefunc)


    ! set parameters for pairlist
    !
    ! ..
    ! ..skip


    ! set parameters for dynamic variables
    !
    call setup_dynvars(dynvars)


    ! set parameters for dynamics
    !
    call setup_dynamics_peer_0(ctrl_data%dyn_info, &
                               ctrl_data%res_info, dynamics)


    ! set parameters for ensemble
    !
    ! ..skip


    ! set parameters for constraint
    !
    call setup_constraints_peer_0(ctrl_data%cons_info, &
                                     par, enefunc, constraints)


    return

  end subroutine setup_md_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_boundary_peer_0
  !> @brief        peer routine of sp_boundary_mod :: setup_boundary()
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_boundary_peer_0(bound_info, table, pairlistdist, watermodel,&
                            ensemble, rigid_bond, boundary_rst, boundary)

    ! formal arguments
    type(s_boundary_info),   intent(in)    :: bound_info
    logical,                 intent(in)    :: table
    real(wp),                intent(in)    :: pairlistdist
    character(*),            intent(in)    :: watermodel
    integer,                 intent(in)    :: ensemble
    logical,                 intent(in)    :: rigid_bond
    type(s_boundary),        intent(in)    :: boundary_rst
    type(s_boundary),        intent(inout) :: boundary


    call init_boundary(boundary)

    if (.not.((bound_info%domain_x == 2 .and. &
               bound_info%domain_y == 1 .and. &
               bound_info%domain_z == 1) .or. &
              (bound_info%domain_x == 2 .and. &
               bound_info%domain_y == 2 .and. &
               bound_info%domain_z == 1))) then

      if (bound_info%domain_x == 1 .or. &
          bound_info%domain_y == 1 .or. &
          bound_info%domain_z == 1 ) then
        call error_msg('Setup_Boundary> other than (2,1,1)/(2,2,1), '//&
                       'domain[x,y,z] should be larger than 1')
      end if

    end if

    boundary%type           = bound_info%type
    boundary%origin_x       = bound_info%origin_x
    boundary%origin_y       = bound_info%origin_y
    boundary%origin_z       = bound_info%origin_z
    boundary%box_size_x     = bound_info%pbc_info%box_size_x
    boundary%box_size_y     = bound_info%pbc_info%box_size_y
    boundary%box_size_z     = bound_info%pbc_info%box_size_z
    boundary%box_size_x_ref = boundary%box_size_x
    boundary%box_size_y_ref = boundary%box_size_y
    boundary%box_size_z_ref = boundary%box_size_z
    
    if (boundary_rst%box_size_x /= 0.0_wp) then

      boundary%box_size_x     = boundary_rst%box_size_x
      boundary%box_size_y     = boundary_rst%box_size_y
      boundary%box_size_z     = boundary_rst%box_size_z
      boundary%box_size_x_ref = boundary%box_size_x
      boundary%box_size_y_ref = boundary%box_size_y
      boundary%box_size_z_ref = boundary%box_size_z

    end if

    call setup_processor_number(bound_info, &
                                table, pairlistdist, watermodel, &
                                ensemble, rigid_bond, boundary)

    call setup_boundary_cell   (table, pairlistdist, watermodel, &
                                ensemble, rigid_bond, boundary)

    return

  end subroutine setup_boundary_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_domain_peer_0
  !> @brief        peer routine of sp_domain_mod::setup_domain()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_domain_peer_0(ene_info, con_info, &
                                 par, boundary, enefunc, constraints)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_cons_info),       intent(in)    :: con_info
    type(s_par),             intent(in)    :: par
    type(s_boundary),        intent(in)    :: boundary
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_constraints),     intent(inout) :: constraints


    ! initialize structure informations
    !
    call init_enefunc(enefunc)
    call init_constraints(constraints)

    enefunc%table%table       = ene_info%table
    enefunc%table%water_model = ene_info%water_model

    constraints%rigid_bond    = con_info%rigid_bond
    constraints%fast_water    = con_info%fast_water
    constraints%water_model   = con_info%water_model


    if (constraints%rigid_bond) then

      call setup_solute_and_water_peer_0(constraints%water_model, enefunc)

      call setup_hbond_group_peer_0     (par, enefunc, constraints)

    else if (enefunc%table%table .and. &
             enefunc%table%water_model(1:4) == 'TIP3') then

      call setup_solute_and_water_peer_0(enefunc%table%water_model, enefunc)

    else

      ! nothing to do

    end if


    return

  end subroutine setup_domain_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_solute_and_water_peer_0
  !> @brief        peer routine of sp_domain_mod::setup_solute_and_water()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_solute_and_water_peer_0(water_model, enefunc)

    ! formal arguments
    character(*),            intent(in)    :: water_model
    type(s_enefunc),         intent(inout) :: enefunc
    
    ! local variables
    integer                  :: i, i1, i2, i3, io, ih
    integer                  :: nwat, nsol, mole_no, tmp_mole_no
    character(4)             :: a1, a2, a3
    logical                  :: lexist


    ! s_enefunc%table
    !   num_water
    !   num_solute
    !   water_list
    !   solute_list
    !   solute_list_inv
    !
   
    nwat = 0
    tmp_mole_no = -1

    do i = 1, cm_num_bonds

      i1 = cm_bond_list(1,i)
      i2 = cm_bond_list(2,i)

      if (cm_residue_name(i1) == water_model .and. &
          cm_residue_name(i2) == water_model) then

        mole_no = cm_molecule_no(i1)
        if (mole_no /= tmp_mole_no) then
          nwat = nwat + 1
          tmp_mole_no = mole_no
        end if

      end if

    end do

    write(MsgOut,'(a,i9)') &
         'Setup_Solute_And_Water_Peer_0> # of water molecule    : ', nwat

    call cm_create_water_list(nwat, lexist)

    if (.not. lexist) then

      nwat = 0
      tmp_mole_no = -1

      do i = 1, cm_num_angles
        
        i1 = cm_angl_list(1,i)
        i2 = cm_angl_list(2,i)
        i3 = cm_angl_list(3,i)

        if (cm_residue_name(i1) == water_model .and. &
            cm_residue_name(i2) == water_model .and. &
            cm_residue_name(i3) == water_model) then

          mole_no = cm_molecule_no(i1)
          if (mole_no /= tmp_mole_no) then
            nwat = nwat + 1
            tmp_mole_no = mole_no

            a1 = cm_atom_name(i1)
            a2 = cm_atom_name(i2)
            a3 = cm_atom_name(i3)

            if (a1(1:1)=='H' .and. a2(1:1)=='O' .and. a3(1:1)=='H') then
              call cm_set_water_list(1, nwat, i2)

              if (i1 > i3) then
                call cm_set_water_list(2, nwat, i3)
                call cm_set_water_list(3, nwat, i1)
              else
                call cm_set_water_list(2, nwat, i1)
                call cm_set_water_list(3, nwat, i3)
              end if

            end if

          end if
        end if
      end do
    end if

    call cm_create_solute(cm_num_atoms, lexist)

    if (.not. lexist) then

      do i = 1, cm_num_atoms
        call cm_set_solute(i, .true.)
      end do

      do i = 1, cm_num_water_list
        i1 = cm_water_list(1, i)
        i2 = cm_water_list(2, i)
        i3 = cm_water_list(3, i)
        call cm_set_solute(i1, .false.)
        call cm_set_solute(i2, .false.)
        call cm_set_solute(i3, .false.)
      end do

    end if

    nsol = 0
    do i = 1, cm_num_atoms
      if (cm_solute(i)) &
        nsol = nsol + 1
    end do

    write(MsgOut,'(a,i9)') &
         'Setup_Solute_And_Water_Peer_0> # of solute            : ', nsol

    call cm_create_solute_list    (nsol,         lexist)
    call cm_create_solute_list_inv(cm_num_atoms, lexist)

    if (.not. lexist) then

      nsol = 0

      do i = 1, cm_num_atoms
        call cm_set_solute_list_inv(i, 0)
      end do

      do i = 1, cm_num_atoms
        if (cm_solute(i)) then
          nsol = nsol + 1
          call cm_set_solute_list    (nsol, i)
          call cm_set_solute_list_inv(i,    nsol)
        end if
      end do

    end if

    write(MsgOut,'(a)') ' '

    ! set water oxygen and hydrogen
    !
    io = cm_water_list(1,1)
    ih = cm_water_list(2,1)

    enefunc%table%atom_cls_no_O = cm_atom_cls_no(io)
    enefunc%table%atom_cls_no_H = cm_atom_cls_no(ih)
    enefunc%table%charge_O      = cm_charge(io)
    enefunc%table%charge_H      = cm_charge(ih)
    enefunc%table%mass_O        = cm_mass(io)
    enefunc%table%mass_H        = cm_mass(ih)

    return

  end subroutine setup_solute_and_water_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_hbond_group_peer_0
  !> @brief        peer routine of sp_domain_mod::setup_hbond_group()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_hbond_group_peer_0(par, enefunc, constraints)

    ! formal arguments
    type(s_par),             intent(in)    :: par
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_constraints),     intent(inout) :: constraints

    ! local variables
    integer                  :: i, j, k, l, i1, i2, idu
    integer                  :: natom, nbond, nbond_p, nsolute, nhmax
    character(6)             :: ci1, ci2
    logical                  :: lexist


    nsolute = cm_num_solute_list
    nbond_p = par%num_bonds

    nbond = cm_num_bonds
    natom = cm_num_atoms

    call cm_create_duplicate(natom, lexist)
    call cm_create_H_index  (natom, lexist)

    if (.not. lexist) then

      do i = 1, natom
        call cm_set_duplicate(i, 0)
        do j = 1, 8
          call cm_set_H_index(j, i, 0)
        end do
      end do

      do i = 1, nbond

        i1 = cm_bond_list(1,i)
        i2 = cm_bond_list(2,i)

        if (cm_solute_list_inv(i1) /= 0 .and. &
            cm_solute_list_inv(i2) /= 0) then

          ci1 = cm_atom_name(i1)
          ci2 = cm_atom_name(i2)

          if (ci1(1:1) == 'H' .or. ci2(1:1) == 'H') then

            if (ci1(1:1) == 'H') then
              idu = cm_duplicate(i2) + 1
              call cm_set_duplicate(i2, idu)
              call cm_set_H_index  (idu, i2, i1)
            else
              idu = cm_duplicate(i1) + 1
              call cm_set_duplicate(i1, idu)
              call cm_set_H_index  (idu, i1, i2)
            end if

          end if

        end if
      end do

    end if

    ! count XHn group for each number n
    !
    constraints%nh(1:8) = 0
    do i = 1, nsolute
      i1 = cm_solute_list(i)
      if (cm_duplicate(i1) == 1) then
        constraints%nh(1) = constraints%nh(1) + 1
      else if (cm_duplicate(i1) == 2) then
        constraints%nh(2) = constraints%nh(2) + 1
      else if (cm_duplicate(i1) == 3) then
        constraints%nh(3) = constraints%nh(3) + 1
      else if (cm_duplicate(i1) == 4) then
        constraints%nh(4) = constraints%nh(4) + 1
      else if (cm_duplicate(i1) == 5) then
        constraints%nh(5) = constraints%nh(5) + 1
      else if (cm_duplicate(i1) == 6) then
        constraints%nh(6) = constraints%nh(6) + 1
      else if (cm_duplicate(i1) == 7) then
        constraints%nh(7) = constraints%nh(7) + 1
      else if (cm_duplicate(i1) == 8) then
        constraints%nh(8) = constraints%nh(8) + 1
      else if (cm_duplicate(i1) >= 8) then
        call error_msg("Bond(>8) for one atom is not considered")
      end if
    end do

    constraints%connect  = 0
    if (constraints%nh(1) /= 0) constraints%connect = 1
    if (constraints%nh(2) /= 0) constraints%connect = 2
    if (constraints%nh(3) /= 0) constraints%connect = 3
    if (constraints%nh(4) /= 0) constraints%connect = 4
    if (constraints%nh(5) /= 0) constraints%connect = 5
    if (constraints%nh(6) /= 0) constraints%connect = 6
    if (constraints%nh(7) /= 0) constraints%connect = 7
    if (constraints%nh(8) /= 0) constraints%connect = 8

    nhmax = max(constraints%nh(1),constraints%nh(2),constraints%nh(3), &
                constraints%nh(4),constraints%nh(5),constraints%nh(6), &
                constraints%nh(7),constraints%nh(8))
    call alloc_constraints(constraints, ConstraintsBondGroup, nhmax)

    ! Make a list of XHn
    !
    constraints%nh(1:8) = 0
    do i = 1, nsolute

      i1 = cm_solute_list(i)

      if (cm_duplicate(i1) == 1) then
    
        constraints%nh(1) = constraints%nh(1) + 1
        constraints%H_Group(1, constraints%nh(1), 1)= i1
        constraints%H_Group(2, constraints%nh(1), 1)= cm_H_index(1, i1)

      else if (cm_duplicate(i1) == 2) then

        constraints%nh(2) = constraints%nh(2) + 1
        constraints%H_Group(1, constraints%nh(2), 2)= i1       
        constraints%H_Group(2, constraints%nh(2), 2)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(2), 2)= cm_H_index(2, i1)

      else if (cm_duplicate(i1) == 3) then

        constraints%nh(3) = constraints%nh(3) + 1
        constraints%H_Group(1, constraints%nh(3), 3)= i1       
        constraints%H_Group(2, constraints%nh(3), 3)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(3), 3)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(3), 3)= cm_H_index(3, i1)

      
      else if (cm_duplicate(i1) == 4) then

        constraints%nh(4) = constraints%nh(4) + 1
        constraints%H_Group(1, constraints%nh(4), 4)= i1       
        constraints%H_Group(2, constraints%nh(4), 4)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(4), 4)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(4), 4)= cm_H_index(3, i1)
        constraints%H_Group(5, constraints%nh(4), 4)= cm_H_index(4, i1)


      else if (cm_duplicate(i1) == 5) then

        constraints%nh(5) = constraints%nh(5) + 1
        constraints%H_Group(1, constraints%nh(5), 5)= i1       
        constraints%H_Group(2, constraints%nh(5), 5)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(5), 5)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(5), 5)= cm_H_index(3, i1)
        constraints%H_Group(5, constraints%nh(5), 5)= cm_H_index(4, i1)
        constraints%H_Group(6, constraints%nh(5), 5)= cm_H_index(5, i1)


      else if (cm_duplicate(i1) == 6) then

        constraints%nh(6) = constraints%nh(6) + 1
        constraints%H_Group(1, constraints%nh(6), 6)= i1       
        constraints%H_Group(2, constraints%nh(6), 6)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(6), 6)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(6), 6)= cm_H_index(3, i1)
        constraints%H_Group(5, constraints%nh(6), 6)= cm_H_index(4, i1)
        constraints%H_Group(6, constraints%nh(6), 6)= cm_H_index(5, i1)
        constraints%H_Group(7, constraints%nh(6), 6)= cm_H_index(6, i1)


      else if (cm_duplicate(i1) == 7) then

        constraints%nh(7) = constraints%nh(7) + 1
        constraints%H_Group(1, constraints%nh(7), 7)= i1       
        constraints%H_Group(2, constraints%nh(7), 7)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(7), 7)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(7), 7)= cm_H_index(3, i1)
        constraints%H_Group(5, constraints%nh(7), 7)= cm_H_index(4, i1)
        constraints%H_Group(6, constraints%nh(7), 7)= cm_H_index(5, i1)
        constraints%H_Group(7, constraints%nh(7), 7)= cm_H_index(6, i1)
        constraints%H_Group(8, constraints%nh(7), 7)= cm_H_index(7, i1)


      else if (cm_duplicate(i1) == 8) then

        constraints%nh(8) = constraints%nh(8) + 1
        constraints%H_Group(1, constraints%nh(8), 8)= i1       
        constraints%H_Group(2, constraints%nh(8), 8)= cm_H_index(1, i1)
        constraints%H_Group(3, constraints%nh(8), 8)= cm_H_index(2, i1)
        constraints%H_Group(4, constraints%nh(8), 8)= cm_H_index(3, i1)
        constraints%H_Group(5, constraints%nh(8), 8)= cm_H_index(4, i1)
        constraints%H_Group(6, constraints%nh(8), 8)= cm_H_index(5, i1)
        constraints%H_Group(7, constraints%nh(8), 8)= cm_H_index(6, i1)
        constraints%H_Group(8, constraints%nh(8), 8)= cm_H_index(7, i1)
        constraints%H_Group(9, constraints%nh(8), 8)= cm_H_index(8, i1)

      end if
  
    end do


    ! mass and bond distance
    !

    constraints%num_bonds = 0
    do i = 1, constraints%connect
      do j = 1, constraints%nh(i)
        ci1 = cm_atom_cls_name(constraints%H_Group(1,j,i))
        do k = 1, i  
          ci2 = cm_atom_cls_name(constraints%H_Group(k+1,j,i))

          do l = 1, nbond_p

            if ((ci1 == par%bond_atom_cls(1,l) .and. &
                 ci2 == par%bond_atom_cls(2,l)) .or. &
                (ci2 == par%bond_atom_cls(1,l) .and. &
                 ci1 == par%bond_atom_cls(2,l))) then
              constraints%num_bonds = constraints%num_bonds + 1
              constraints%bond_dist(k,j,i) = par%bond_dist_min(l)
              exit
            end if
          end do

        end do
      end do
    end do

    return

  end subroutine setup_hbond_group_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_restraints_peer_0
  !> @brief        peer routine of sp_restraints_mod::setup_restraints()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_restraints_peer_0(res_info, sel_info, restraints)

    ! formal arguments
    type(s_res_info),        intent(in)    :: res_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_restraints),      intent(inout) :: restraints

    ! local variables
    type(s_selatoms)         :: selatoms
    integer                  :: i, nfunc, ngroup
    logical                  :: lexist


    !write(MsgOut,'(a)') '  Setup_Restraints_Peer_0> '


    ! select atoms
    do i = 1, size(sel_info%groups)
      call select_atom_peer(sel_info%groups(i), selatoms)
      call cm_create_selatoms(selatoms, lexist)
    end do
    call dealloc_selatoms(selatoms)

    ! copy s_res_info to s_restraint

    restraints%nfunctions = res_info%nfunctions
    restraints%num_groups = size(sel_info%groups)

    nfunc  = restraints%nfunctions
    ngroup = restraints%num_groups

    call alloc_restraints(restraints, RestraintsFunc,  nfunc)
    call alloc_restraints(restraints, RestraintsGroup, ngroup)

    restraints%function     (1:nfunc) = res_info%function     (1:nfunc)
    restraints%exponent     (1:nfunc) = res_info%exponent     (1:nfunc)
    restraints%constant     (1:nfunc) = res_info%constant     (1:nfunc)
    restraints%select_index (1:nfunc) = res_info%select_index (1:nfunc)
    restraints%reference    (1:nfunc) = res_info%reference    (1:nfunc)
    restraints%exponent_dist(1:nfunc) = res_info%exponent_dist(1:nfunc)
    restraints%weight_dist  (1:nfunc) = res_info%weight_dist  (1:nfunc)
    restraints%group       (1:ngroup) = sel_info%groups      (1:ngroup)

    restraints%restraint_flag         = (restraints%nfunctions > 0)

    restraints%max_atoms = 0
    do i = 1, size(sel_info%groups)
      restraints%max_atoms = max(restraints%max_atoms, cm_num_selatoms(i))
    end do

    return

  end subroutine setup_restraints_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc_peer_0
  !> @brief        peer routine of sp_enefunc_mod::define_enefunc()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc_peer_0(ene_info,   &
                                   par,        &
                                   restraints, &
                                   enefunc)

    ! formal arguments
    type(s_ene_info),     intent(in)    :: ene_info
    type(s_par),          intent(in)    :: par
    type(s_restraints),   intent(in)    :: restraints
    type(s_enefunc),      intent(inout) :: enefunc


    ! nonbonded
    !
    call setup_enefunc_nonb_peer_0(ene_info, par, enefunc)


    ! lookup table
    !
    call setup_enefunc_table_peer_0(ene_info, enefunc)


    ! restraint
    !
    call setup_enefunc_restraints_peer_0(restraints, enefunc)

    return

  end subroutine define_enefunc_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_nonb_peer_0
  !> @brief        peer routine of sp_enefunc_charmm_mod::setup_enefunc_nonb()
  !! @authors      NT
  !! @param[in]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_nonb_peer_0(ene_info, par, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_par),             intent(in)    :: par
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: eps14, rmin14, eps, rmin
    integer                  :: i, j, nonb_p


    !write(MsgOut,'(a)') '  Setup_Enefunc_Nonb_Peer_0> '

    nonb_p = par%num_atom_cls

    ! setup cutoff distance etc.
    !
    enefunc%num_atom_cls = nonb_p

    ! set lennard-jones parameters
    !
    call alloc_enefunc(enefunc, EneFuncNbon, nonb_p)

    do i = 1, nonb_p

      lj_coef(1,i) = par%nonb_eps (i)
      lj_coef(2,i) = par%nonb_rmin(i)

      do j = 1, nonb_p

        ! combination rule
        !
        eps14  = sqrt(par%nonb_eps_14(i) * par%nonb_eps_14(j))
        rmin14 = par%nonb_rmin_14(i) + par%nonb_rmin_14(j)
        eps    = sqrt(par%nonb_eps(i) * par%nonb_eps(j))
        rmin   = par%nonb_rmin(i) + par%nonb_rmin(j)

        ! set parameters
        !
        enefunc%nb14_lj12(i,j) = eps14 * (rmin14 ** 12)
        enefunc%nb14_lj6(i,j)  = 2.0_wp * eps14 * (rmin14 ** 6)
        enefunc%nonb_lj12(i,j) = eps * (rmin ** 12)
        enefunc%nonb_lj6(i,j)  = 2.0_wp * eps * (rmin ** 6)

      end do
    end do

    return

  end subroutine setup_enefunc_nonb_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_table_peer_0
  !> @brief        peer routine of sp_enefunc_table_mod::setup_enefunc_table()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_table_peer_0(ene_info, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_enefunc),         intent(inout) :: enefunc


    !write(MsgOut,'(a)') '  Setup_Enefunc_Table_Peer_0> '

    enefunc%table%table_order = ene_info%table_order
    enefunc%table%water_model = ene_info%water_model

    return

  end subroutine setup_enefunc_table_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_restraints_peer_0
  !> @brief        peer routine of 
  !!                      sp_enefunc_restraints_mod::setup_enefunc_restraints()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_restraints_peer_0(restraints, enefunc)

    ! formal arguments
    type(s_restraints),      intent(in)    :: restraints
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: i, j
    integer                  :: nconst, nref


    enefunc%restraint = restraints%restraint_flag

    if (.not. restraints%restraint_flag) return

    ! setup group
    !
    call setup_enefunc_rest_group_peer_0(restraints, enefunc)


    ! setup func
    !
    call setup_enefunc_rest_func_peer_0(restraints, enefunc)


    ! setup about domain
    !
    ! ..skip


    write(MsgOut,'(A)') 'Setup_Enefunc_Restraint_Peer_0> Setup restraint groups'

    do i = 1, enefunc%num_restraintgroups

      write(MsgOut,'(A,I5,3A)') &
           ' group = ', i, ', "', trim(restraints%group(i)),'"'
      write(MsgOut,'(A,I5)')    &
           ' # of atoms = ', cm_num_selatoms(i)
      write(MsgOut,'(A)') &
           ' atomlist: '

      do j = 1, cm_num_selatoms(i)

        write(MsgOut,'(i5,$)') cm_selatoms(j,i)
        if (mod(j, 10) == 0 .and. j /= cm_num_selatoms(i)) then
          write(MsgOut,'(A)') ''
        end if

      end do
      write(MsgOut,'(A)') ''

    end do

    write(MsgOut,'(A)') &
                     'Setup_Enefunc_Restraint_Peer_0> Setup restraint functions'

    do i = 1, enefunc%num_restraintfuncs

      write(MsgOut,'(A,I5,A,I5)') &
        ' func  = ', i, ' kind  = ', enefunc%restraint_kind(i)

      ! summary for positional restraint
      !
      if (enefunc%restraint_kind(i) == RestraintsFuncPOSI) then
        write(MsgOut,'(A,4F8.3)') &
          ' const(total, x, y, z) = ', enefunc%restraint_const(1:4,i) 
      else
        write(MsgOut,'(A,2F8.3)') &
          ' const(Lower, Upper) = ',   enefunc%restraint_const(1:2,i)
        write(MsgOut,'(A,2F8.3)') &
          ' ref  (Lower, Upper) = ',   enefunc%restraint_ref(1:2,i)
      end if

      write(MsgOut,'(A,I5)') ' exponend of function = ',                &
                             enefunc%restraint_exponent_func(i)

      ! summary for distance restraint
      !
      if (enefunc%restraint_kind(i) == RestraintsFuncDIST .or.          &
          enefunc%restraint_kind(i) == RestraintsFuncDISTCOM) then

        call error_msg('Setup_Enefunc_Restraint_Peer_0> '//&
                       'DIHED restraint is not supported in parallel I/O')

      ! summary for angle restraint
      !
      else if (enefunc%restraint_kind(i) == RestraintsFuncANGLE .or.    &
               enefunc%restraint_kind(i) == RestraintsFuncANGLECOM ) then

        call error_msg('Setup_Enefunc_Restraint_Peer_0> '//&
                       'ANGLE restraint is not supported in parallel I/O')

      ! summary for dihedral angle restraint
      !
      else if (enefunc%restraint_kind(i) == RestraintsFuncDIHED  .or.   &
               enefunc%restraint_kind(i) == RestraintsFuncDIHEDCOM ) then

        call error_msg('Setup_Enefunc_Restraint_Peer_0> '//&
                       'DIHED restraint is not supported in parallel I/O')

      ! summary for positinal and RMSD restraint
      !
      else 

        write(MsgOut,'(A,I5)') ' # of groups  = ', enefunc%restraint_funcgrp(i)
        write(MsgOut,'(" grouplist: ",$)') 
        do j = 1, enefunc%restraint_funcgrp(i) 
          write(MsgOut,'(i3,$)') enefunc%restraint_grouplist(j,i)
          if (mod(j,20) == 0 .and. j /= enefunc%restraint_funcgrp(i) )  &
            write(MsgOut,'(A)') ''
        end do
        write(MsgOut,'(A)') ''

      end if
      write(MsgOut,'(A)') ''

    end do

    return

  end subroutine setup_enefunc_restraints_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_rest_group_peer_0
  !> @brief        peer routine of 
  !!                     sp_enefunc_restraints_mod::setup_enefunc_rest_group()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_rest_group_peer_0(restraints, enefunc)

    ! formal arguments
    type(s_restraints),      intent(in)    :: restraints
    type(s_enefunc),         intent(inout) :: enefunc


    enefunc%num_restraintgroups = restraints%num_groups

    return

  end subroutine setup_enefunc_rest_group_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      setup_enefunc_rest_func_peer_0
  !> @brief        peer routine of 
  !!                      sp_enefunc_restraints_mod::setup_enefunc_rest_func()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_rest_func_peer_0(restraints, enefunc) 

    ! formal arguments
    type(s_restraints),      intent(in)     :: restraints
    type(s_enefunc),         intent(inout)  :: enefunc

    ! local variables
    integer                  :: i, ndata
    integer                  :: num_funcs, max_grp
    integer,     allocatable :: num_indexes(:)


    ! get total number of groups in index
    !
    enefunc%num_restraintfuncs = restraints%nfunctions
    num_funcs = enefunc%num_restraintfuncs
    allocate(num_indexes(1:num_funcs))

    do i = 1, num_funcs
      ndata = split_num(restraints%select_index(i))

      select case(restraints%function(i))
      case(RestraintsFuncDIST:RestraintsFuncDISTCOM)
        if (ndata <  2 .or. mod(ndata,2) /= 0) &
          call error_msg( &
            'Setup_Enefunc_Rest_Func_Peer_0> Error in index for DIST')

      case(RestraintsFuncANGLE:RestraintsFuncANGLECOM)
        if (ndata /= 3) &
          call error_msg( &
            'Setup_Enefunc_Rest_Func_Peer_0> Error in index for ANGLE')

      case(RestraintsFuncDIHED:RestraintsFuncDIHEDCOM)
        if (ndata /= 4) &
          call error_msg( &
            'Setup_Enefunc_Rest_Func_Peer_0> Error in index for DIHE')

      case default
        if (ndata /= 1) & 
          call error_msg( &
            'Setup_Enefunc_Rest_Func_Peer_0> Error in index for POSI or RMSD')

      end select

      num_indexes(i) = ndata
    end do

    ! get maximum number of groups among all 'index's
    !
    max_grp = maxval(num_indexes(1:num_funcs))
    enefunc%max_restraint_numgrps = max_grp

    call alloc_enefunc(enefunc, EneFuncReff, num_funcs, max_grp)


    ! set parameters 
    !
    do i = 1, num_funcs

      enefunc%restraint_funcgrp(i) = num_indexes(i)
      enefunc%restraint_kind(i)    = restraints%function(i)

      ! setup group list
      !
      call setup_restraint_grouplist(i, restraints, enefunc)

      ! setup exponent_func
      !
      call setup_restraint_exponent_func(i, restraints, enefunc)

      ! setup force constants
      !
      call setup_restraint_constants(i, restraints, enefunc)

      ! setup reference   (except for POSI)
      !
      call setup_restraint_reference(i, restraints, enefunc)

      ! setup exponent_dist (for DIST only)
      !
      call setup_restraint_exponent_dist(i, restraints, enefunc)

      ! setup weight_dist   (for DIST only)
      !
      call setup_restraint_weight_dist(i, restraints, enefunc)

    end do

    ! deallocate local array
    !
    deallocate(num_indexes)

    return

  end subroutine setup_enefunc_rest_func_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      setup_enefunc_rest_refcoord_peer_0
  !> @brief        peer routine of 
  !!                  sp_enefunc_restraints_mod::setup_enefunc_rest_refcoord()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_rest_refcoord_peer_0(enefunc) 

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc


    ! nothing to do

    return

  end subroutine setup_enefunc_rest_refcoord_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_dynamics_peer_0
  !> @brief        peer routine of sp_dynamics_mod::setup_dynamics()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_dynamics_peer_0(dyn_info, res_info, dynamics)

    ! formal arguments
    type(s_dyn_info),        intent(in)    :: dyn_info
    type(s_res_info),        intent(in)    :: res_info
    type(s_dynamics),        intent(inout) :: dynamics

    ! local variables
    integer                  :: i, dec_num_deg_freedom


    ! initialize variables in dynamics
    !
    call init_dynamics(dynamics)


    ! setup variables
    !
    dynamics%iseed = dyn_info%iseed


    ! update number of degrees of freedom
    !
    dec_num_deg_freedom = 3

    do i = 1, res_info%nfunctions
      if (res_info%function(i) == RestraintsFuncPOSI) &
        dec_num_deg_freedom = 0
    end do

    cm_num_deg_freedom = cm_num_atoms * 3 - dec_num_deg_freedom

    return

  end subroutine setup_dynamics_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_constraints_peer_0
  !> @brief        peer routine of sp_constraints_mod::setup_constraints()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_constraints_peer_0(cons_info, par, enefunc, constraints)

    ! formal arguments
    type(s_cons_info),       intent(in)    :: cons_info
    type(s_par),             intent(in)    :: par
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_constraints),     intent(inout) :: constraints

    integer                  :: iw1, iw2


    constraints%rigid_bond      = cons_info%rigid_bond
    constraints%fast_bond       = cons_info%fast_bond
    constraints%fast_water      = cons_info%fast_water
    constraints%shake_iteration = cons_info%shake_iteration
    constraints%shake_tolerance = cons_info%shake_tolerance
    constraints%lincs_order     = cons_info%lincs_order
    constraints%lincs_iteration = cons_info%lincs_iteration
    constraints%water_model     = cons_info%water_model

    if (constraints%rigid_bond) then

      ! setup SETTLE
      !
      if (constraints%fast_water) then

        call setup_fast_water_peer_0(par, enefunc, constraints)

        ! update number of degrees of freedom
        !
        cm_num_deg_freedom = cm_num_deg_freedom - 3 * cm_num_water_list

      end if


      ! setup SHAKE and RATTLE
      !
      call setup_rigid_bond_peer_0(par, enefunc, constraints)


      ! update number of degrees of freedom
      !
      cm_num_deg_freedom = cm_num_deg_freedom - constraints%num_bonds

    end if

    return

  end subroutine setup_constraints_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_fast_water_peer_0
  !> @brief        peer routine of sp_constraints_mod::setup_fast_water()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_fast_water_peer_0(par, enefunc, constraints)

    ! formal arguments
    type(s_par),             intent(in)    :: par
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_constraints),     intent(inout) :: constraints

    ! local variables
    integer                  :: i, list(3)
    character(6)             :: ci1, ci2, ci3


    list(1) = cm_water_list(1,1)
    list(2) = cm_water_list(2,1)
    list(3) = cm_water_list(3,1)
    constraints%water_massO = cm_mass(list(1))
    constraints%water_massH = cm_mass(list(2))

    ci1 = cm_atom_cls_name(list(1))
    ci2 = cm_atom_cls_name(list(2))
    ci3 = cm_atom_cls_name(list(3))

    do i = 1, par%num_bonds
      if ((ci1 == par%bond_atom_cls(1,i) .and. &
           ci2 == par%bond_atom_cls(2,i)) .or. &
          (ci1 == par%bond_atom_cls(2,i) .and. &
           ci2 == par%bond_atom_cls(1,i))) then
        constraints%water_rOH = par%bond_dist_min(i)
        exit
      end if
    end do

    do i = 1, par%num_bonds
      if (ci2 == par%bond_atom_cls(1,i) .and. &
          ci3 == par%bond_atom_cls(2,i)) then
        constraints%water_rHH = par%bond_dist_min(i)
        exit
      end if
    end do

    return

  end subroutine setup_fast_water_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_rigid_bond_peer_0
  !> @brief        peer routine of sp_constraints_mod::setup_rigid_bond()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_rigid_bond_peer_0(par, enefunc, constraints)

    ! formal arguments
    type(s_par),             intent(in)    :: par
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_constraints),     intent(inout) :: constraints


    ! nothing to do

    return

  end subroutine setup_rigid_bond_peer_0

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_md_peer
  !> @brief        peer routine of sp_setup_spdyn_mod::setup_md()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_md_peer(local_my_rank, &
                           ctrl_data,     &
                           par,           &
                           domain_index,  &
                           boundary,      &
                           enefunc,       &
                           constraints,   &
                           domain)

    ! formal arguments
    integer,                 intent(in)    :: local_my_rank
    type(s_ctrl_data),       intent(in)    :: ctrl_data
    type(s_par),             intent(in)    :: par
    type(s_domain_index),    intent(in)    :: domain_index
    type(s_boundary),        intent(in)    :: boundary
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_constraints),     intent(inout) :: constraints
    type(s_domain),          intent(inout) :: domain


    !write(MsgOut,'(a)') 'Setup_Md_Peer> '


    ! set parameters for domain 
    !
    call setup_domain_peer(ctrl_data%ene_info,  &
                           ctrl_data%cons_info, &
                           local_my_rank, domain_index, boundary, &
                           enefunc, constraints, domain)


    ! setup enefunc in each domain
    !
    call define_enefunc_peer(par, ctrl_data%ene_info, &
                             domain_index, constraints, domain, enefunc)


    !write(MsgOut,'(a)') ' '

    domain%num_deg_freedom  = cm_num_deg_freedom

    return

  end subroutine setup_md_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_domain_peer
  !> @brief        peer routine of sp_domain_mod::setup_domain()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_domain_peer(ene_info,      &
                               con_info,      &
                               local_my_rank, &
                               domain_index,  &
                               boundary,      &
                               enefunc,       &
                               constraints,   &
                               domain)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_cons_info),       intent(in)    :: con_info
    integer,                 intent(in)    :: local_my_rank
    type(s_domain_index),    intent(in)    :: domain_index
    type(s_boundary),        intent(in)    :: boundary
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_constraints),     intent(inout) :: constraints
    type(s_domain),          intent(inout) :: domain

    ! local variables
    integer                  :: i, j, k, cell(3)
    integer                  :: icel_local, icel
    integer                  :: ncel_local, ncel_bound, ncel_all


    ! initialize structure informations
    !
    call init_domain(domain)

    domain%num_atom_all = cm_num_atoms

    enefunc%table%table       = ene_info%table
    enefunc%table%water_model = ene_info%water_model

    constraints%rigid_bond    = con_info%rigid_bond
    constraints%fast_water    = con_info%fast_water
    constraints%water_model   = con_info%water_model


    ! assign the rank of each dimension from my_rank
    !

    !$omp critical

        my_world_rank = local_my_rank
        my_city_rank  = local_my_rank

        call setup_processor_rank(boundary, domain, cell)

    !$omp end critical


    ! decide cell capacity (max**) for memory allocation
    !
    call setup_cell_capacity(boundary, domain)


    ! memory allocaltion of maps connecting local to global cell indices
    !
    ncel_local = domain%num_cell_local
    ncel_bound = domain%num_cell_boundary
    ncel_all   = ncel_local + ncel_bound

    call alloc_domain(domain, DomainCellGlobal, cell(1),cell(2),cell(3))
    call alloc_domain(domain, DomainCellLocal,    ncel_local, 1, 1)
    call alloc_domain(domain, DomainCellLocBou,   ncel_all,   1, 1)
    call alloc_domain(domain, DomainCellBoundary, ncel_bound, 1, 1)
    call alloc_domain(domain, DomainCellPair,     ncel_all,   1, 1)


    ! assign global<->local mapping of cell indexa
    !
    icel_local = 0
    do i = domain%cell_start(3), domain%cell_end(3)
      do j = domain%cell_start(2), domain%cell_end(2)
        do k = domain%cell_start(1), domain%cell_end(1)
          icel_local = icel_local + 1
          icel = k + (j-1)*cell(1) + (i-1)*cell(1)*cell(2)
          domain%cell_g2l(icel) = icel_local
          domain%cell_l2g(icel_local) = icel
          domain%cell_l2gx(icel_local) = k
          domain%cell_l2gy(icel_local) = j
          domain%cell_l2gz(icel_local) = i
          domain%cell_l2gx_orig(icel_local) = k
          domain%cell_l2gy_orig(icel_local) = j
          domain%cell_l2gz_orig(icel_local) = i
          domain%cell_gxyz2l(k,j,i) = icel_local
        end do
      end do
    end do


    ! assigin each boundary cell
    !
    call setup_cell_boundary(cell, boundary%num_domain, domain)


    ! assign of atom maps connecting global local to global atom indices
    !
    call alloc_domain(domain, DomainDynvar, ncel_all,  1, 1)
    call alloc_id_g2l(domain, cm_num_atoms)


    if (constraints%rigid_bond) then

      call setup_atom_by_HBond_peer(domain_index, boundary, constraints, domain)

    else if (enefunc%table%table .and. &
             enefunc%table%water_model(1:4) == 'TIP3') then

      call setup_atom_by_table_peer(domain_index, boundary, domain)

    else

      call setup_atom_peer         (domain_index, boundary, domain)

    end if


    !$omp critical
    
        my_world_rank = local_my_rank
        my_city_rank  = local_my_rank

        call setup_domain_interaction(boundary, domain)

    !$omp end critical

    return

  end subroutine setup_domain_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_atom_by_HBond_peer
  !> @brief        peer routine of sp_domain_mod::setup_atom_by_HBond_peer()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_atom_by_HBond_peer(domain_index, &
                                      boundary,     &
                                      constraints,  &
                                      domain)

    ! formal arguments
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_boundary),     target, intent(in)    :: boundary
    type(s_constraints),  target, intent(inout) :: constraints
    type(s_domain),       target, intent(inout) :: domain

    ! local variable
    real(wp)                      :: x_shift, y_shift, z_shift
    real(wp)                      :: move(3), origin(3)
    integer                       :: i, j, icx, icy, icz, icel
    integer                       :: isolute, iwater, ih1, ih2
    integer                       :: iisolute, iiwater
    integer                       :: icel_local
    integer                       :: ncel, ncel_local
    character(4)                  :: ci1

    real(wp),             pointer :: bsize_x, bsize_y, bsize_z
    real(wp),             pointer :: csize_x, csize_y, csize_z
    real(wp),             pointer :: HGr_bond_dist(:,:,:,:)
    integer,              pointer :: iwater_list(:), isolute_list(:)
    integer,              pointer :: ncel_x, ncel_y, ncel_z
    integer,              pointer :: cell_g2l(:), cell_g2b(:)
    integer,              pointer :: natom(:), nsolute(:), nwater(:)
    integer,              pointer :: solute_list(:,:), water_list(:,:,:)
    integer,              pointer :: No_HGr(:), HGr_local(:,:)
    integer,              pointer :: HGr_bond_list(:,:,:,:)
    integer,              pointer :: patm, psol, pwat, pnoh, phgl


    !write(MsgOut,'(a)') '    Setup_Atom_By_HBond_Peer> '

    ncel          = domain%num_cell_local + domain%num_cell_boundary
    ncel_local    = domain%num_cell_local

    call alloc_constraints(constraints, ConstraintsDomainBond, ncel, &
                           constraints%connect)

    iwater_list   => domain_index%iwater_list
    isolute_list  => domain_index%isolute_list

    bsize_x       => boundary%box_size_x
    bsize_y       => boundary%box_size_y
    bsize_z       => boundary%box_size_z
    ncel_x        => boundary%num_cells_x
    ncel_y        => boundary%num_cells_y
    ncel_z        => boundary%num_cells_z
    csize_x       => boundary%cell_size_x
    csize_y       => boundary%cell_size_y
    csize_z       => boundary%cell_size_z

    cell_g2l      => domain%cell_g2l
    cell_g2b      => domain%cell_g2b
    natom         => domain%num_atom
    nsolute       => domain%num_solute
    nwater        => domain%num_water
    solute_list   => domain%solute_list
    water_list    => domain%water_list

    No_HGr        => constraints%No_HGr
    HGr_local     => constraints%HGr_local
    HGr_bond_list => constraints%HGr_bond_list
    HGr_bond_dist => constraints%HGr_bond_dist

    origin(1)     = boundary%origin_x
    origin(2)     = boundary%origin_y
    origin(3)     = boundary%origin_z

    natom      (                       1:ncel) = 0
    nsolute    (                       1:ncel) = 0
    nwater     (                       1:ncel) = 0
    solute_list(          1:MaxAtom,   1:ncel) = 0
    water_list (1:3,      1:MaxWater,  1:ncel) = 0
    No_HGr     (                       1:ncel) = 0
    HGr_local  (1:constraints%connect, 1:ncel) = 0


    ! solute atoms (not bonded to hydrogen) in each domain
    !
    do iisolute = 1, domain_index%nsolute_list

      isolute = isolute_list(iisolute)

      i = cm_solute_list(isolute)

      ci1 = cm_atom_name(i)

      if (cm_duplicate(i) == 0 .and. ci1(1:1) /= 'H') then

        !coordinate shifted against the origin
        !
        x_shift = cm_atom_coord(1, i) - boundary%origin_x
        y_shift = cm_atom_coord(2, i) - boundary%origin_y
        z_shift = cm_atom_coord(3, i) - boundary%origin_z

        !coordinate shifted to the first quadrant and set into the boundary box
        !
        move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
        move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
        move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
        x_shift = x_shift + move(1)
        y_shift = y_shift + move(2)
        z_shift = z_shift + move(3)

        !assign which cell
        !
        icx = int(x_shift/csize_x)
        icy = int(y_shift/csize_y)
        icz = int(z_shift/csize_z)
        if (icx == ncel_x) icx = icx - 1
        if (icy == ncel_y) icy = icy - 1
        if (icz == ncel_z) icz = icz - 1
        icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

        ! atoms inside the domain
        !
        if (cell_g2l(icel) /= 0) then

          ! local cell index
          !
          icel_local = cell_g2l(icel)

          patm => natom  (icel_local)
          psol => nsolute(icel_local)
          pnoh => No_HGr (icel_local)

          ! local_count : total number of atoms in each cell
          !
          patm = patm + 1
          psol = psol + 1
          pnoh = pnoh + 1

          solute_list(psol,icel_local) = patm

          call molecule_to_domain_peer(move, origin, i, &
                                       domain, icel_local, patm)

        ! atoms in the boundary
        !
        else if (cell_g2b(icel) /= 0) then
  
          ! local cell index
          !
          icel_local = cell_g2b(icel) + ncel_local

          patm => natom  (icel_local)
          psol => nsolute(icel_local)
          pnoh => No_HGr (icel_local)

          ! local_count : total number of atoms in each cell
          !
          patm = patm + 1
          psol = psol + 1
          pnoh = pnoh + 1

          solute_list(psol,icel_local) = patm

          call molecule_to_domain_peer(move, origin, i, &
                                       domain, icel_local, patm)

        end if 

      end if

    end do

    ! Hydrogen bonding group
    !
    do isolute = 1, constraints%connect

      do j = 1, constraints%nh(isolute)

        i = constraints%H_Group(1, j, isolute)
        x_shift = cm_atom_coord(1, i) - boundary%origin_x
        y_shift = cm_atom_coord(2, i) - boundary%origin_y
        z_shift = cm_atom_coord(3, i) - boundary%origin_z

        move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
        move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
        move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
        x_shift = x_shift + move(1)
        y_shift = y_shift + move(2)
        z_shift = z_shift + move(3)

        !assign which cell
        !
        icx = int(x_shift/csize_x)
        icy = int(y_shift/csize_y)
        icz = int(z_shift/csize_z)
        if (icx == ncel_x) icx = icx - 1
        if (icy == ncel_y) icy = icy - 1
        if (icz == ncel_z) icz = icz - 1
        icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

        ! atoms inside the domain
        !
        if (cell_g2l(icel) /= 0) then
  
          icel_local = cell_g2l(icel)

          patm => natom  (icel_local)
          psol => nsolute(icel_local)
          phgl => HGr_local(isolute,icel_local)
  
          ! nonhydrogen atom
          !
          patm = patm + 1
          psol = psol + 1
          phgl = phgl + 1

          solute_list(psol,icel_local) = patm
          HGr_bond_list(1,phgl,isolute,icel_local) = patm 

          call molecule_to_domain_peer(move, origin, i, &
                                       domain, icel_local, patm)

          do ih1 = 1, isolute

            i = constraints%H_Group(ih1+1, j, isolute)

            patm = patm + 1
            psol = psol + 1

            solute_list(psol,icel_local) = patm
            HGr_bond_list(ih1+1,phgl,isolute,icel_local) = patm 
            HGr_bond_dist(ih1+1,phgl,isolute,icel_local) = &
                                         constraints%bond_dist(ih1,j,isolute)

            call molecule_to_domain_peer(move, origin, i, &
                                         domain, icel_local, patm)

          end do

        else if (cell_g2b(icel) /= 0) then

          icel_local = cell_g2b(icel) + ncel_local

          patm => natom  (icel_local)
          psol => nsolute(icel_local)
          phgl => HGr_local(isolute,icel_local)

          ! hydrogen atoms
          !
          patm = patm + 1
          psol = psol + 1
          phgl = phgl + 1

          solute_list(psol,icel_local) = patm
          HGr_bond_list(1,phgl,isolute,icel_local) = patm 

          call molecule_to_domain_peer(move, origin, i, &
                                       domain, icel_local, patm)

          do ih1 = 1, isolute

            i = constraints%H_Group(ih1+1,j,isolute)

            patm = patm + 1
            psol = psol + 1

            solute_list(psol,icel_local) = patm
            HGr_bond_list(ih1+1,phgl,isolute,icel_local) = patm 

            call molecule_to_domain_peer(move, origin, i, &
                                         domain, icel_local, patm)

          end do

        end if

      end do

    end do

    ! water atoms in each domain
    !
    do iiwater = 1, domain_index%nwater_list

      iwater = iwater_list(iiwater)

      i   = cm_water_list(1,iwater)
      ih1 = cm_water_list(2,iwater)
      ih2 = cm_water_list(3,iwater)

      !coordinate shifted against the origin
      !
      x_shift = cm_atom_coord(1, i) - boundary%origin_x
      y_shift = cm_atom_coord(2, i) - boundary%origin_y
      z_shift = cm_atom_coord(3, i) - boundary%origin_z

      !coordinate shifted to the first quadrant and set into the boundary box
      move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
      move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
      move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
      x_shift = x_shift + move(1)
      y_shift = y_shift + move(2)
      z_shift = z_shift + move(3)

      !assign which cell
      !
      icx = int(x_shift/csize_x)
      icy = int(y_shift/csize_y)
      icz = int(z_shift/csize_z)
      if (icx == ncel_x) icx = icx - 1
      if (icy == ncel_y) icy = icy - 1
      if (icz == ncel_z) icz = icz - 1
      icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

      ! atoms inside the domain
      !
      if (cell_g2l(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2l(icel)

        patm => natom (icel_local)
        pwat => nwater(icel_local)

        pwat = pwat + 1

        ! oxygen atoms
        !
        patm = patm + 1
        water_list(1,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

        ! first hydrogen atoms
        !
        patm = patm + 1
        water_list(2,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih1, &
                                     domain, icel_local, patm)

        ! second hydrogen atoms
        !
        patm = patm + 1
        water_list(3,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih2, &
                                     domain, icel_local, patm)

      ! atoms in the boundary
      !
      else if (cell_g2b(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2b(icel) + ncel_local

        patm => natom (icel_local)
        pwat => nwater(icel_local)

        pwat = pwat + 1

        ! oxygen atoms
        !
        patm = patm + 1
        water_list(1,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

        ! first hydrogen atoms
        !
        patm = patm + 1
        water_list(2,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih1, &
                                     domain, icel_local, patm)


        ! second hydrogen atoms
        !
        patm = patm + 1
        water_list(3,pwat,icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih2, &
                                     domain, icel_local, patm)

      end if

    end do

    domain%num_atom_t0(1:ncel) = natom(1:ncel)

    return

  end subroutine setup_atom_by_HBond_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_atom_by_table_peer
  !> @brief        peer routine of sp_domain_mod::setup_atom_table()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_atom_by_table_peer(domain_index, &
                                      boundary,     &
                                      domain)

    ! formal arguments
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_boundary),     target, intent(in)    :: boundary
    type(s_domain),       target, intent(inout) :: domain

    ! local variables
    real(wp)                      :: x_shift, y_shift, z_shift
    real(wp)                      :: move(3), origin(3)
    integer                       :: i, ih1, ih2
    integer                       :: icx, icy, icz, icel, icel_local
    integer                       :: isolute, iwater
    integer                       :: iisolute, iiwater
    integer                       :: ncel, ncel_local

    real(wp),             pointer :: bsize_x, bsize_y, bsize_z
    real(wp),             pointer :: csize_x, csize_y, csize_z
    integer,              pointer :: iwater_list(:), isolute_list(:)
    integer,              pointer :: ncel_x, ncel_y, ncel_z
    integer,              pointer :: cell_g2l(:), cell_g2b(:)
    integer,              pointer :: natom(:), nsolute(:), nwater(:)
    integer,              pointer :: solute_list(:,:), water_list(:,:,:)
    integer,              pointer :: patm, psol, pwat


    !write(MsgOut,'(a)') '    Setup_Atom_By_Table_Peer> '

    iwater_list  => domain_index%iwater_list
    isolute_list => domain_index%isolute_list

    bsize_x      => boundary%box_size_x
    bsize_y      => boundary%box_size_y
    bsize_z      => boundary%box_size_z
    ncel_x       => boundary%num_cells_x
    ncel_y       => boundary%num_cells_y
    ncel_z       => boundary%num_cells_z
    csize_x      => boundary%cell_size_x
    csize_y      => boundary%cell_size_y
    csize_z      => boundary%cell_size_z

    cell_g2l     => domain%cell_g2l
    cell_g2b     => domain%cell_g2b
    natom        => domain%num_atom
    nsolute      => domain%num_solute
    nwater       => domain%num_water
    solute_list  => domain%solute_list
    water_list   => domain%water_list

    origin(1)    = boundary%origin_x
    origin(2)    = boundary%origin_y
    origin(3)    = boundary%origin_z

    ncel         = domain%num_cell_local + domain%num_cell_boundary
    ncel_local   = domain%num_cell_local

    natom      (                 1:ncel) = 0
    nsolute    (                 1:ncel) = 0
    nwater     (                 1:ncel) = 0
    solute_list(     1:MaxAtom,  1:ncel) = 0
    water_list (1:3, 1:MaxWater, 1:ncel) = 0


    ! solute atoms in each domain
    !
    do iisolute = 1, domain_index%nsolute_list

      isolute = isolute_list(iisolute)

      i = cm_solute_list(isolute)

      ! coordinate shifted against the origin
      !
      x_shift = cm_atom_coord(1,i) - boundary%origin_x
      y_shift = cm_atom_coord(2,i) - boundary%origin_y
      z_shift = cm_atom_coord(3,i) - boundary%origin_z

      ! coordinate shifted to the first quadrant and set into the boundary box
      !
      move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
      move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
      move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
      x_shift = x_shift + move(1)
      y_shift = y_shift + move(2)
      z_shift = z_shift + move(3)

      ! assign which cell
      !
      icx = int(x_shift/csize_x)
      icy = int(y_shift/csize_y)
      icz = int(z_shift/csize_z)
      if (icx == ncel_x) icx = icx - 1
      if (icy == ncel_y) icy = icy - 1
      if (icz == ncel_z) icz = icz - 1
      icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

      ! atoms inside the domain
      !
      if (cell_g2l(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2l(icel)

        patm => natom  (icel_local)
        psol => nsolute(icel_local)

        ! local_count : total number of atoms in each cell
        !
        patm = patm + 1
        psol = psol + 1

        solute_list(psol, icel_local) = patm
        solute_list(patm, icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

      ! atoms in the boundary
      !
      else if (cell_g2b(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2b(icel) + ncel_local

        patm => natom  (icel_local)
        psol => nsolute(icel_local)

        ! local_count : total number of atoms in each cell
        !
        patm = patm + 1
        psol = psol + 1

        solute_list(psol, icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

      end if

    end do


    ! water atoms in each domain
    !
    do iiwater = 1, domain_index%nwater_list

      iwater = iwater_list(iiwater)

      i   = cm_water_list(1, iwater)
      ih1 = cm_water_list(2, iwater)
      ih2 = cm_water_list(3, iwater)

      ! coordinate shifted against the origin
      !
      x_shift = cm_atom_coord(1, i) - boundary%origin_x
      y_shift = cm_atom_coord(2, i) - boundary%origin_y
      z_shift = cm_atom_coord(3, i) - boundary%origin_z

      ! coordinate shifted to the first quadrant and set into the boundary box
      !
      move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
      move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
      move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
      x_shift = x_shift + move(1)
      y_shift = y_shift + move(2)
      z_shift = z_shift + move(3)

      !assign which cell
      !
      icx = int(x_shift/csize_x)
      icy = int(y_shift/csize_y)
      icz = int(z_shift/csize_z)
      if (icx == ncel_x) icx = icx - 1
      if (icy == ncel_y) icy = icy - 1
      if (icz == ncel_z) icz = icz - 1
      icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

      ! atoms inside the domain
      !
      if (cell_g2l(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2l(icel)

        patm => natom (icel_local)
        pwat => nwater(icel_local)

        pwat = pwat + 1

        ! oxygen atoms
        !
        patm = patm + 1
        water_list(1, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

        ! first hydrogen atoms
        !
        patm = patm + 1
        water_list(2, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih1, &
                                     domain, icel_local, patm)

        ! second hydrogen atoms
        !
        patm = patm + 1
        water_list(3, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih2, &
                                     domain, icel_local, patm)

      ! atoms in the boundary
      !
      else if (cell_g2b(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2b(icel) + ncel_local

        patm => natom (icel_local)
        pwat => nwater(icel_local)

        pwat = pwat + 1

        ! oxygen atoms
        !
        patm = patm + 1
        water_list(1, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, patm)

        ! first hydrogen atoms
        !
        patm = patm + 1
        water_list(2, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih1, &
                                     domain, icel_local, patm)

        ! second hydrogen atoms
        !
        patm = patm + 1
        water_list(3, pwat, icel_local) = patm

        call molecule_to_domain_peer(move, origin, ih2, &
                                     domain, icel_local, patm)

      end if

    end do

    domain%num_atom_t0(1:ncel) = natom(1:ncel)

    return

  end subroutine setup_atom_by_table_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_atom_peer
  !> @brief        peer routine of sp_domain_mod::setup_atom()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_atom_peer(domain_index, &
                             boundary,     &
                             domain)

    ! formal arguments
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_boundary),     target, intent(in)    :: boundary
    type(s_domain),       target, intent(inout) :: domain

    ! local variables
    real(wp)                      :: x_shift, y_shift, z_shift
    real(wp)                      :: move(3), origin(3)
    integer                       :: i, iatom
    integer                       :: icx, icy, icz, icel, icel_local
    integer                       :: ncel, ncel_local

    real(wp),             pointer :: bsize_x, bsize_y, bsize_z
    real(wp),             pointer :: csize_x, csize_y, csize_z
    integer,              pointer :: iatom_list(:)
    integer,              pointer :: ncel_x, ncel_y, ncel_z
    integer,              pointer :: cell_g2l(:), cell_g2b(:)
    integer,              pointer :: natom(:)


    !write(MsgOut,'(a)') '    Setup_Atom_Peer> '

    iatom_list   => domain_index%iatom_list

    bsize_x      => boundary%box_size_x
    bsize_y      => boundary%box_size_y
    bsize_z      => boundary%box_size_z
    ncel_x       => boundary%num_cells_x
    ncel_y       => boundary%num_cells_y
    ncel_z       => boundary%num_cells_z
    csize_x      => boundary%cell_size_x
    csize_y      => boundary%cell_size_y
    csize_z      => boundary%cell_size_z

    cell_g2l     => domain%cell_g2l
    cell_g2b     => domain%cell_g2b
    natom        => domain%num_atom

    origin(1)    = boundary%origin_x
    origin(2)    = boundary%origin_y
    origin(3)    = boundary%origin_z

    ncel         = domain%num_cell_local + domain%num_cell_boundary
    ncel_local   = domain%num_cell_local

    natom      (                 1:ncel) = 0


    do iatom = 1, domain_index%natom_list

      i = iatom_list(iatom)

      !coordinate shifted against the origin
      !
      x_shift = cm_atom_coord(1,i) - boundary%origin_x
      y_shift = cm_atom_coord(2,i) - boundary%origin_y
      z_shift = cm_atom_coord(3,i) - boundary%origin_z

      !coordinate shifted to the first quadrant and set into the boundary box
      !
      move(1) = bsize_x*0.5_wp - bsize_x*anint(x_shift/bsize_x)
      move(2) = bsize_y*0.5_wp - bsize_y*anint(y_shift/bsize_y)
      move(3) = bsize_z*0.5_wp - bsize_z*anint(z_shift/bsize_z)
      x_shift = x_shift + move(1)
      y_shift = y_shift + move(2)
      z_shift = z_shift + move(3)

      !assign which cell
      !
      icx = int(x_shift/csize_x)
      icy = int(y_shift/csize_y)
      icz = int(z_shift/csize_z)
      if (icx == ncel_x) icx = icx - 1
      if (icy == ncel_y) icy = icy - 1
      if (icz == ncel_z) icz = icz - 1
      icel = 1 + icx + icy*ncel_x + icz*ncel_x*ncel_y

      ! atoms inside the domain
      !
      if (cell_g2l(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2l(icel)

        natom(icel_local) = natom(icel_local) + 1

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, natom(icel_local))


      ! atoms in a boundary
      !
      else if (cell_g2b(icel) /= 0) then

        ! local cell index
        !
        icel_local = cell_g2b(icel) + ncel_local

        natom(icel_local) = natom(icel_local) + 1

        call molecule_to_domain_peer(move, origin, i, &
                                     domain, icel_local, natom(icel_local))

      end if

    end do

    domain%num_atom_t0(1:ncel) = natom(1:ncel)

    return

  end subroutine setup_atom_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    molecule_to_domain_peer
  !> @brief        peer routine of sp_domain_mod::molecule_to_domain_peer()
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine molecule_to_domain_peer(move, origin, iatom, &
                                     domain, icel, icel_atom)

    ! formal arguments
    real(wp),                 intent(in)    :: move(3)
    real(wp),                 intent(in)    :: origin(3)
    integer,                  intent(in)    :: iatom
    type(s_domain),   target, intent(inout) :: domain
    integer,                  intent(in)    :: icel
    integer,                  intent(in)    :: icel_atom

    ! local variables
    real(wp),         pointer :: coord_local (:,:,:)
    real(wp),         pointer :: ref_local   (:,:,:)
    real(wp),         pointer :: vel_local   (:,:,:)
    real(wp),         pointer :: charge_local(:,:)
    real(wp),         pointer :: mass_local  (:,:)
    real(wp),         pointer :: trans       (:,:,:)
    integer,          pointer :: class_local (:,:)
    integer,          pointer :: id_l2g      (:,:)
    integer,          pointer :: id_g2l      (:,:)
    

    id_g2l       => domain%id_g2l
    id_l2g       => domain%id_l2g
    coord_local  => domain%coord
    ref_local    => domain%coord_ref
    vel_local    => domain%velocity
    charge_local => domain%charge
    mass_local   => domain%mass
    class_local  => domain%atom_cls_no
    trans        => domain%trans_vec


    id_l2g(icel_atom,icel) = iatom
    id_g2l(1,iatom) = icel
    id_g2l(2,iatom) = icel_atom

    coord_local(1,icel_atom,icel) = cm_atom_coord(1,iatom)
    coord_local(2,icel_atom,icel) = cm_atom_coord(2,iatom)
    coord_local(3,icel_atom,icel) = cm_atom_coord(3,iatom)
    ref_local  (1,icel_atom,icel) = cm_atom_coord(1,iatom)
    ref_local  (2,icel_atom,icel) = cm_atom_coord(2,iatom)
    ref_local  (3,icel_atom,icel) = cm_atom_coord(3,iatom)
    vel_local  (1,icel_atom,icel) = cm_atom_velocity(1,iatom)
    vel_local  (2,icel_atom,icel) = cm_atom_velocity(2,iatom)
    vel_local  (3,icel_atom,icel) = cm_atom_velocity(3,iatom)
    charge_local (icel_atom,icel) = cm_charge(iatom)
    mass_local   (icel_atom,icel) = cm_mass(iatom)
    class_local  (icel_atom,icel) = cm_atom_cls_no(iatom)
    trans      (1,icel_atom,icel) = move(1) - origin(1)
    trans      (2,icel_atom,icel) = move(2) - origin(2)
    trans      (3,icel_atom,icel) = move(3) - origin(3)

    return

  end subroutine molecule_to_domain_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_enefunc_peer
  !> @brief        peer routine of sp_enefunc_mod::define_enefunc()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_enefunc_peer(par,          &
                                 ene_info,     &
                                 domain_index, &
                                 constraints,  &
                                 domain,       &
                                 enefunc)

    ! formal arguments
    type(s_par),          intent(in)    :: par
    type(s_ene_info),     intent(in)    :: ene_info
    type(s_domain_index), intent(in)    :: domain_index
    type(s_constraints),  intent(in)    :: constraints
    type(s_domain),       intent(in)    :: domain
    type(s_enefunc),      intent(inout) :: enefunc

    ! local variables
    integer               :: ncel, ncelb


    !write(MsgOut,'(a)') '  Define_Enefunc_Peer> '


    ! base
    !
    ncel  = domain%num_cell_local
    ncelb = domain%num_cell_local + domain%num_cell_boundary

    call alloc_enefunc(enefunc, EneFuncBase,     ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncBond,     ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncAngl,     ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncDihe,     ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncImpr,     ncel, ncel)
    call alloc_enefunc(enefunc, EneFuncBondCell, ncel, ncelb)


    if (.not. constraints%rigid_bond) then

      ! bond
      !
      call setup_enefunc_bond_peer(par, domain_index, domain, enefunc)


      ! angle
      !
      call setup_enefunc_angl_peer(par, domain_index, domain, enefunc)

    else

      ! bond
      !
      call setup_enefunc_bond_constraint_peer( &
                                        par, domain_index, domain, enefunc)

      ! angle
      !
      call setup_enefunc_angl_constraint_peer( &
                                        par, domain_index, domain, enefunc)

    end if


    ! dihedral
    !
    call setup_enefunc_dihe_peer(par, domain_index, domain, enefunc)


    ! improper
    !
    call setup_enefunc_impr_peer(par, domain_index, domain, enefunc)


    ! cmap
    !
    call setup_enefunc_cmap_peer(par, domain_index, domain, enefunc, &
                                        ene_info)

    ! nonbonded
    !
    ! call setup_enefunc_nonb() ...skip


    ! restraint
    !
    call setup_enefunc_restraints_peer(domain_index, domain, enefunc)

    return

  end subroutine define_enefunc_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_peer
  !> @brief        peer routine of sp_enefunc_mod::setup_enefunc_bond()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_peer(par, domain_index, domain, enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variable
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: nbond_p, i1, i2
    integer                       :: nbond_lst, ibond
    character(6)                  :: ci1, ci2

    real(wp),             pointer :: force(:,:), dist(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: bond(:), list(:,:,:)
    integer,              pointer :: ibond_lst(:)

    
    !write(MsgOut,'(a)') '    Setup_Enefunc_Bond_Peer> '

    ibond_lst => domain_index%ibond_list

    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair

    bond      => enefunc%num_bond
    list      => enefunc%bond_list
    force     => enefunc%bond_force_const
    dist      => enefunc%bond_dist_min

    nbond_lst = domain_index%nbond_list
    nbond_p   = par%num_bonds


    do ibond = 1, nbond_lst

      i = ibond_lst(ibond)

      i1  = cm_bond_list(1, i)
      i2  = cm_bond_list(2, i)
      ci1 = cm_atom_cls_name(i1)
      ci2 = cm_atom_cls_name(i2)
    
      icel1 = domain%id_g2l(1, i1)
      icel2 = domain%id_g2l(1, i2)

      ! Check if it is in my domain
      !
      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1, icel2)

        if (icel_local > 0 .and. icel_local <= ncel) then
 
          do j = 1, nbond_p
            if ((ci1 == par%bond_atom_cls(1, j) .and.  &
                 ci2 == par%bond_atom_cls(2, j)) .or.  &
                (ci1 == par%bond_atom_cls(2, j) .and.  &
                 ci2 == par%bond_atom_cls(1, j))) then
              bond (icel_local) = bond(icel_local) + 1
              list (1, bond(icel_local), icel_local) = i1
              list (2, bond(icel_local), icel_local) = i2
              force(bond(icel_local), icel_local) = par%bond_force_const(j) 
              dist (bond(icel_local), icel_local) = par%bond_dist_min(j)
              exit
            end if
          end do
 
          if (j == nbond_p + 1) &
            write(MsgOut,*) &
                 'Setup_Enefunc_Bond_Peer> not found BOND: [', &
                 ci1, ']-[', ci2, '] in parameter file. (ERROR)'

        end if
      end if

    end do

    return

  end subroutine setup_enefunc_bond_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl_peer
  !> @brief        peer routine of sp_enefunc_mod::setup_enefunc_angl()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl_peer(par, domain_index, domain, enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variables
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: nangl_p
    integer                       :: list(3)
    integer                       :: nangl_lst, iangl
    character(6)                  :: ci1, ci2, ci3

    real(wp),             pointer :: force(:,:), theta(:,:)
    real(wp),             pointer :: ubforce(:,:), ubrmin(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: angle(:), alist(:,:,:)
    integer,              pointer :: iangl_list(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Angl_Peer> '

    iangl_list => domain_index%iangl_list

    ncel       => domain%num_cell_local
    cell_pair  => domain%cell_pair

    angle      => enefunc%num_angle
    alist      => enefunc%angle_list
    force      => enefunc%angle_force_const
    theta      => enefunc%angle_theta_min
    ubforce    => enefunc%urey_force_const
    ubrmin     => enefunc%urey_rmin

    nangl_lst  = domain_index%nangl_list
    nangl_p    = par%num_angles


    do iangl = 1, nangl_lst

      i = iangl_list(iangl)

      list(1) = cm_angl_list(1, i)
      list(2) = cm_angl_list(2, i)
      list(3) = cm_angl_list(3, i)
      ci1 = cm_atom_cls_name(list(1))
      ci2 = cm_atom_cls_name(list(2))
      ci3 = cm_atom_cls_name(list(3))

      icel1 = domain%id_g2l(1, list(1))
      icel2 = domain%id_g2l(1, list(3))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1, icel2)
       
        if (icel_local >= 1 .and. icel_local <= ncel) then

          do j = 1, nangl_p
            if ((ci1 == par%angl_atom_cls(1, j) .and. &
                 ci2 == par%angl_atom_cls(2, j) .and. &
                 ci3 == par%angl_atom_cls(3, j)) .or. &
                (ci1 == par%angl_atom_cls(3, j) .and. &
                 ci2 == par%angl_atom_cls(2, j) .and. &
                 ci3 == par%angl_atom_cls(1, j))) then
              angle(icel_local) = angle(icel_local) + 1
              alist(1:3,angle(icel_local),icel_local) = list(1:3)
              force  (angle(icel_local),icel_local) = par%angl_force_const(j)
              theta  (angle(icel_local),icel_local) = par%angl_theta_min(j)*RAD
              ubforce(angle(icel_local),icel_local) = par%angl_ub_force_const(j)
              ubrmin (angle(icel_local),icel_local) = par%angl_ub_rmin(j)
              exit
            end if
          end do

          if (j == nangl_p + 1) &
            write(MsgOut,*) &
              'Setup_Enefunc_Angl_Peer> not found ANGL: [', &
              ci1, ']-[', ci2, ']-[', ci3, '] in parameter file. (ERROR)'

        end if
      end if

    end do

    return

  end subroutine setup_enefunc_angl_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_dihe_peer
  !> @brief        peer routine of sp_enefunc_mod::setup_enefunc_dihe()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_dihe_peer(par, domain_index, domain, enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variables
    integer                       :: ndihe_p
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: nw_found
    integer                       :: list(4)
    integer                       :: ndihe_lst, idihe
    character(6)                  :: ci1, ci2, ci3, ci4

    real(wp),             pointer :: force(:,:), phase(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: dihedral(:), dlist(:,:,:), period(:,:)
    integer,              pointer :: idihe_lst(:)

    logical,          allocatable :: no_wild(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Dihe_Peer> '

    idihe_lst  => domain_index%idihe_list
    
    ncel       => domain%num_cell_local
    cell_pair  => domain%cell_pair

    dihedral   => enefunc%num_dihedral
    dlist      => enefunc%dihe_list
    force      => enefunc%dihe_force_const
    period     => enefunc%dihe_periodicity
    phase      => enefunc%dihe_phase
   
    ndihe_lst  = domain_index%ndihe_list
    ndihe_p    = par%num_dihedrals


    ! check usage of wild card
    !
    allocate(no_wild(ndihe_p))

    do i = 1, ndihe_p

      if ((par%dihe_atom_cls(1,i) /= WildCard) .and. &
          (par%dihe_atom_cls(4,i) /= WildCard)) then
        ! A-B-C-D type
        no_wild(i) = .true.
      else
        ! X-B-C-D type
        no_wild(i) = .false.
      end if

    end do

    ! find number of interactions
    !
    do idihe = 1, ndihe_lst

      i = idihe_lst(idihe)

      list(1) = cm_dihe_list(1, i)
      list(2) = cm_dihe_list(2, i)
      list(3) = cm_dihe_list(3, i)
      list(4) = cm_dihe_list(4, i)
      ci1 = cm_atom_cls_name(list(1))
      ci2 = cm_atom_cls_name(list(2))
      ci3 = cm_atom_cls_name(list(3))
      ci4 = cm_atom_cls_name(list(4))

      icel1 = domain%id_g2l(1, list(1))
      icel2 = domain%id_g2l(1, list(4))


      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          nw_found = 0
          do j = 1, ndihe_p
            if (no_wild(j)) then
              if (((ci1 == par%dihe_atom_cls(1, j)) .and. &
                   (ci2 == par%dihe_atom_cls(2, j)) .and. &
                   (ci3 == par%dihe_atom_cls(3, j)) .and. &
                   (ci4 == par%dihe_atom_cls(4, j))) .or. &
                  ((ci1 == par%dihe_atom_cls(4, j)) .and. &
                   (ci2 == par%dihe_atom_cls(3, j)) .and. &
                   (ci3 == par%dihe_atom_cls(2, j)) .and. &
                   (ci4 == par%dihe_atom_cls(1, j)))) then
                nw_found = nw_found + 1
                dihedral(icel_local) = dihedral(icel_local) + 1 
                dlist (1:4,dihedral(icel_local),icel_local) = list(1:4)
                force (dihedral(icel_local),icel_local) =par%dihe_force_const(j)
                period(dihedral(icel_local),icel_local) =par%dihe_periodicity(j)
                phase (dihedral(icel_local),icel_local) =par%dihe_phase(j) * RAD
              end if
            end if
          end do

          if (nw_found == 0) then
            do j = 1, ndihe_p
              if (.not.no_wild(j)) then
                if (((ci2 == par%dihe_atom_cls(2, j)) .and. &
                     (ci3 == par%dihe_atom_cls(3, j))) .or. &
                    ((ci2 == par%dihe_atom_cls(3, j)) .and. &
                     (ci3 == par%dihe_atom_cls(2, j)))) then
                  dihedral(icel_local) = dihedral(icel_local) + 1
                  dlist (1:4,dihedral(icel_local),icel_local) = list(1:4)
                  force (dihedral(icel_local),icel_local) = &
                       par%dihe_force_const(j)
                  period(dihedral(icel_local),icel_local) = &
                       par%dihe_periodicity(j)
                  phase (dihedral(icel_local),icel_local) = &
                       par%dihe_phase(j) * RAD
                end if
              end if
            end do
          end if

        end if
      end if
    end do

    deallocate(no_wild)

    return

  end subroutine setup_enefunc_dihe_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_impr_peer
  !> @brief        peer routine of sp_enefunc_mod::setup_enefunc_impr()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_impr_peer(par, domain_index, domain, enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variables
    integer                       :: nimpr_p
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: list(4)
    integer                       :: nimpr_lst, iimpr
    character(6)                  :: ci1, ci2, ci3, ci4
    logical                       :: no_wild

    real(wp),             pointer :: force(:,:), phase(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: improper(:), ilist(:,:,:)
    integer,              pointer :: iimpr_lst(:)
    integer,          allocatable :: wc_type(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Impr_Peer> '

    iimpr_lst  => domain_index%iimpr_list
    
    ncel       => domain%num_cell_local
    cell_pair  => domain%cell_pair

    improper   => enefunc%num_improper
    ilist      => enefunc%impr_list
    force      => enefunc%impr_force_const
    phase      => enefunc%impr_phase

    nimpr_lst  = domain_index%nimpr_list
    nimpr_p    = par%num_impropers


    ! check usage of wild card
    !
    allocate(wc_type(nimpr_p))

    do i = 1, nimpr_p

      if ((par%impr_atom_cls(1,i) /= WildCard) .and. &
          (par%impr_atom_cls(2,i) /= WildCard) .and. &
          (par%impr_atom_cls(3,i) /= WildCard) .and. &
          (par%impr_atom_cls(4,i) /= WildCard)) then
        ! A-B-C-D type
        wc_type(i) = 0

      else if ((par%impr_atom_cls(1,i) == WildCard) .and. &
               (par%impr_atom_cls(2,i) /= WildCard) .and. &
               (par%impr_atom_cls(3,i) /= WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then
        ! X-B-C-D type
        wc_type(i) = 1

      else if ((par%impr_atom_cls(1,i) == WildCard) .and. &
               (par%impr_atom_cls(2,i) == WildCard) .and. &
               (par%impr_atom_cls(3,i) /= WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then
        ! X-X-C-D type
        wc_type(i) = 2

      else if ((par%impr_atom_cls(1,i) /= WildCard) .and. &
               (par%impr_atom_cls(2,i) == WildCard) .and. &
               (par%impr_atom_cls(3,i) == WildCard) .and. &
               (par%impr_atom_cls(4,i) /= WildCard)) then
        ! A-X-X-D type
        wc_type(i) = 3

      else
        call error_msg('Setup_Enefunc_Impr_Peer> Undefined Wild Card')

      end if

    end do

    ! setup parameters
    !

    do iimpr = 1, nimpr_lst

      i = iimpr_lst(iimpr)

      no_wild = .false.

      list(1) = cm_impr_list(1, i)
      list(2) = cm_impr_list(2, i)
      list(3) = cm_impr_list(3, i)
      list(4) = cm_impr_list(4, i)
      ci1 = cm_atom_cls_name(list(1))
      ci2 = cm_atom_cls_name(list(2))
      ci3 = cm_atom_cls_name(list(3))
      ci4 = cm_atom_cls_name(list(4))

      icel1 = domain%id_g2l(1, list(1))
      icel2 = domain%id_g2l(1, list(4))

      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = cell_pair(icel1,icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          ! A-B-C-D type
          !
          do j = 1, nimpr_p
            if (wc_type(j) == 0) then
              if (((ci1 == par%impr_atom_cls(1, j)) .and. &
                   (ci2 == par%impr_atom_cls(2, j)) .and. &
                   (ci3 == par%impr_atom_cls(3, j)) .and. &
                   (ci4 == par%impr_atom_cls(4, j))) .or. &
                  ((ci1 == par%impr_atom_cls(4, j)) .and. &
                   (ci2 == par%impr_atom_cls(3, j)) .and. &
                   (ci3 == par%impr_atom_cls(2, j)) .and. &
                   (ci4 == par%impr_atom_cls(1, j)))) then

                improper(icel_local) = improper(icel_local) + 1
                ilist(1:4,improper(icel_local), icel_local) = list(1:4)
                force(improper(icel_local),icel_local) = par%impr_force_const(j)
                phase(improper(icel_local),icel_local) = par%impr_phase(j) * RAD
                no_wild = .true.
                exit

              end if
            end if
          end do

          ! X-B-C-D type
          !
          if (.not.no_wild) then
            do j = 1, nimpr_p
              if (wc_type(j) == 1) then
                if (((ci2 == par%impr_atom_cls(2, j)) .and. &
                     (ci3 == par%impr_atom_cls(3, j)) .and. &
                     (ci4 == par%impr_atom_cls(4, j))) .or. &
                    ((ci2 == par%impr_atom_cls(4, j)) .and. &
                     (ci3 == par%impr_atom_cls(3, j)) .and. &
                     (ci4 == par%impr_atom_cls(2, j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local), icel_local) = list(1:4)
                  force(improper(icel_local),icel_local)=par%impr_force_const(j)
                  phase(improper(icel_local),icel_local)=par%impr_phase(j)*RAD
                  no_wild = .true.
                  exit

                end if
              end if
            end do
          end if

          ! X-X-C-D type
          !
          if (.not.no_wild) then
            do j = 1, nimpr_p
              if (wc_type(j) == 2) then
                if (((ci3 == par%impr_atom_cls(3, j)) .and. &
                     (ci4 == par%impr_atom_cls(4, j))) .or. &
                    ((ci3 == par%impr_atom_cls(4, j)) .and. &
                     (ci4 == par%impr_atom_cls(3, j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local), icel_local) = list(1:4)
                  force(improper(icel_local),icel_local)=par%impr_force_const(j)
                  phase(improper(icel_local),icel_local)=par%impr_phase(j)*RAD
                  no_wild = .true.
                  exit

                end if
              end if
            end do
          end if

          ! A-X-X-D type
          !
          if (.not.no_wild) then
            do j = 1, nimpr_p
              if (wc_type(j) == 3) then
                if (((ci1 == par%impr_atom_cls(1, j)) .and. &
                     (ci4 == par%impr_atom_cls(4, j))) .or. &
                    ((ci1 == par%impr_atom_cls(4, j)) .and. &
                     (ci4 == par%impr_atom_cls(1, j)))) then

                  improper(icel_local) = improper(icel_local) + 1
                  ilist(1:4,improper(icel_local), icel_local) = list(1:4)
                  force(improper(icel_local),icel_local)=par%impr_force_const(j)
                  phase(improper(icel_local),icel_local)=par%impr_phase(j)*RAD
                  no_wild = .true.
                  exit

                end if
              end if
            end do
          end if

          if (.not.no_wild) &
            write(MsgOut,*) &
              'Setup_Enefunc_Impr_Peer> Unknown IMPR type. [', &
              ci1, ']-[', ci2, ']-[', ci3, ']-[', ci4, '] (ERROR)'

        end if
      end if

    end do

    deallocate(wc_type)

    return

  end subroutine setup_enefunc_impr_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_cmap_peer
  !> @brief        peer routine of sp_enefunc_mod::setup_enefunc_cmap()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_cmap_peer(par, domain_index, &
                                     domain, enefunc, ene_info)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc
    type(s_ene_info),             intent(in)    :: ene_info

    ! local variables
    integer                       :: ncmap_p, ngrid0
    integer                       :: list(2), icel1, icel2, icel_local
    integer                       :: flag_cmap_type, alloc_stat, dealloc_stat
    integer                       :: i, j, k, l, ityp
    integer                       :: ncmap_lst, icmap
    character(6)                  :: ci1, ci2, ci3, ci4, ci5, ci6, ci7, ci8
    logical                       :: periodic

    real(wp),             pointer :: ccoef(:,:,:,:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: ncmap(:), clist(:,:,:), ctype(:,:), cres(:)
    integer,              pointer :: icmap_lst(:)

    real(wp),         allocatable :: c_ij(:,:,:,:)      ! cmap coeffs


    !write(MsgOut,'(a)') '    Setup_Enefunc_Cmap_Peer> '

    icmap_lst  => domain_index%icmap_list

    ncel       => domain%num_cell_local
    cell_pair  => domain%cell_pair

    ! If 'periodic' is .true.,
    !   then cubic spline with periodic (in dihedral-angle space) boundary
    !   will be applied.
    ! If 'periodic' is .false.,
    !   then natural cubic spline without periodic boudnary
    !   will be applied to expanded cross-term maps.
    !   This is similar with CHARMM's source code.
    !
    periodic   = ene_info%cmap_pspline
    ncmap_lst  = domain_index%ncmap_list
    ncmap_p    = par%num_cmaps

    !  begin
    !
    ngrid0 = 0
    do i = 1, ncmap_p
      ngrid0 = max(ngrid0, par%cmap_resolution(i))
    end do

    call alloc_enefunc(enefunc, EneFuncCmap, ncel, ngrid0, ncmap_p)

    alloc_stat = 0
    allocate(c_ij(4,4,ngrid0,ngrid0), stat = alloc_stat)
    if (alloc_stat /= 0) &
      call error_msg_alloc

    ncmap      => enefunc%num_cmap
    clist      => enefunc%cmap_list
    ctype      => enefunc%cmap_type
    cres       => enefunc%cmap_resolution
    ccoef      => enefunc%cmap_coef

    do i = 1, ncmap_p
      cres(i) = par%cmap_resolution(i)
    end do

    !  derive cmap coefficients by bicubic interpolation
    !
    do ityp = 1, ncmap_p
      if (periodic) then
        call derive_cmap_coefficients_p(ityp, par, c_ij)
      else
        call derive_cmap_coefficients_np(ityp, par, c_ij)
      end if
      do l = 1, ngrid0
        do k = 1, ngrid0 
          do j = 1, 4
            do i = 1, 4
              ccoef(i,j,k,l,ityp) = c_ij(i,j,k,l)
            end do
          end do
        end do
      end do
    end do

    ncmap(1:ncel) = 0

    do icmap = 1, ncmap_lst

      i = icmap_lst(icmap)

      list(1) = cm_cmap_list(1, i)
      list(2) = cm_cmap_list(8, i)
      icel1 = domain%id_g2l(1, list(1))
      icel2 = domain%id_g2l(1, list(2))
      
      if (icel1 /= 0 .and. icel2 /= 0) then

        icel_local = domain%cell_pair(icel1, icel2)

        if (icel_local >= 1 .and. icel_local <= ncel) then

          ! ci* will be atom-type strings
          !
          ci1 = cm_atom_cls_name(cm_cmap_list(1,i))
          ci2 = cm_atom_cls_name(cm_cmap_list(2,i))
          ci3 = cm_atom_cls_name(cm_cmap_list(3,i))
          ci4 = cm_atom_cls_name(cm_cmap_list(4,i))
          ci5 = cm_atom_cls_name(cm_cmap_list(5,i))
          ci6 = cm_atom_cls_name(cm_cmap_list(6,i))
          ci7 = cm_atom_cls_name(cm_cmap_list(7,i))
          ci8 = cm_atom_cls_name(cm_cmap_list(8,i))
          flag_cmap_type = -1

          ! assign cmap type ID to each (psi,phi) pair
          !
          do j = 1, ncmap_p
            if (ci1 == par%cmap_atom_cls(1, j) .and. &
                ci2 == par%cmap_atom_cls(2, j) .and. &
                ci3 == par%cmap_atom_cls(3, j) .and. &
                ci4 == par%cmap_atom_cls(4, j) .and. &
                ci5 == par%cmap_atom_cls(5, j) .and. &
                ci6 == par%cmap_atom_cls(6, j) .and. &
                ci7 == par%cmap_atom_cls(7, j) .and. &
                ci8 == par%cmap_atom_cls(8, j)) then

              ncmap(icel_local) = ncmap(icel_local) + 1
              clist(1, ncmap(icel_local), icel_local) = cm_cmap_list(1, i)
              clist(2, ncmap(icel_local), icel_local) = cm_cmap_list(2, i)
              clist(3, ncmap(icel_local), icel_local) = cm_cmap_list(3, i)
              clist(4, ncmap(icel_local), icel_local) = cm_cmap_list(4, i)
              clist(5, ncmap(icel_local), icel_local) = cm_cmap_list(5, i)
              clist(6, ncmap(icel_local), icel_local) = cm_cmap_list(6, i)
              clist(7, ncmap(icel_local), icel_local) = cm_cmap_list(7, i)
              clist(8, ncmap(icel_local), icel_local) = cm_cmap_list(8, i)
              ctype(ncmap(icel_local), icel_local) = j
              flag_cmap_type = j
              exit
            end if
          end do

          ! if not found, print detail about the error.
          !
          if (flag_cmap_type <= 0) then
            write(MsgOut,*) 'Setup_Enefunc_Cmap_Peer> not found CMAP: '
            write(MsgOut,*) ' [',ci1,']-[',ci2,']-[',ci3,']-[',ci4,']-'
            write(MsgOut,*) ' [',ci5,']-[',ci6,']-[',ci7,']-[',ci8,'] '
            write(MsgOut,*)  'in parameter file. (ERROR)'
          end if

        end if
      end if

    end do

    deallocate(c_ij)

    return

  end subroutine setup_enefunc_cmap_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_bond_constraint_peer
  !> @brief        peer routine of sp_enefunc_mod::
  !!                                          setup_enefunc_bond_constraint()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_bond_constraint_peer(par, domain_index, domain, &
                                                enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variable
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: nbond_p, i1, i2, found
    integer                       :: nbond_lst, ibond
    character(6)                  :: ci1, ci2
    character(4)                  :: a1, a2

    real(wp),             pointer :: force(:,:), dist(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: bond(:), list(:,:,:)
    integer,              pointer :: ibond_lst(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Bond_Constraint_Peer> '

    ibond_lst => domain_index%ibond_list

    ncel      => domain%num_cell_local
    cell_pair => domain%cell_pair

    bond      => enefunc%num_bond
    list      => enefunc%bond_list
    force     => enefunc%bond_force_const
    dist      => enefunc%bond_dist_min

    nbond_lst = domain_index%nbond_list
    nbond_p = par%num_bonds
  
    bond(1:ncel) = 0

    do ibond = 1, nbond_lst
  
      i = ibond_lst(ibond)

      i1 = cm_bond_list(1,i)
      i2 = cm_bond_list(2,i)
      a1 = cm_atom_name(i1)
      a2 = cm_atom_name(i2)
  
      if (a1(1:1) /= 'H' .and. a2(1:1) /= 'H') then

        ci1 = cm_atom_cls_name(i1)
        ci2 = cm_atom_cls_name(i2)
      
        icel1 = domain%id_g2l(1,i1)
        icel2 = domain%id_g2l(1,i2)
  
        ! Check if it is in my domain
        !
        if (icel1 /= 0 .and. icel2 /= 0) then
  
          icel_local = cell_pair(icel1, icel2)
  
          if (icel_local > 0 .and. icel_local <= ncel) then
   
            do j = 1, nbond_p
              if ((ci1 == par%bond_atom_cls(1, j) .and.  &
                   ci2 == par%bond_atom_cls(2, j)) .or.  &
                  (ci1 == par%bond_atom_cls(2, j) .and.  &
                   ci2 == par%bond_atom_cls(1, j))) then
                bond (icel_local) = bond(icel_local) + 1
                list (1, bond(icel_local), icel_local) = i1
                list (2, bond(icel_local), icel_local) = i2
                force(bond(icel_local), icel_local) = par%bond_force_const(j) 
                dist (bond(icel_local), icel_local) = par%bond_dist_min(j)
                exit
              end if
            end do
     
            if (j == nbond_p + 1) &
              write(MsgOut,*) &
                'Setup_Enefunc_Bond_Constraint_Peer> not found BOND: [', &
                ci1, ']-[', ci2, '] in parameter file. (ERROR)'
    
          end if

        end if

      end if
  
    end do

    return

  end subroutine setup_enefunc_bond_constraint_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_angl_constraint_peer
  !> @brief        peer routine of sp_enefunc_mod::
  !!                                          setup_enefunc_angl_constraint()
  !! @authors      NT, JJ
  !! @param[in]    
  !! @date         2012/10/24 (JJ) - removing angle list of water molecules
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_angl_constraint_peer(par, domain_index, &
                                                domain, enefunc)

    ! formal arguments
    type(s_par),                  intent(in)    :: par
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variables
    integer                       :: i, j, icel_local
    integer                       :: icel1, icel2
    integer                       :: nangl_p
    integer                       :: list(3)
    integer                       :: nangl_lst, iangl
    character(6)                  :: ci1, ci2, ci3
    character(4)                  :: ri1, ri2, ri3

    real(wp),             pointer :: force(:,:), theta(:,:)
    real(wp),             pointer :: ubforce(:,:), ubrmin(:,:)
    integer,              pointer :: ncel, cell_pair(:,:)
    integer,              pointer :: angle(:), alist(:,:,:)
    integer,              pointer :: iangl_list(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Angl_Peer> '

    iangl_list => domain_index%iangl_list

    ncel       => domain%num_cell_local
    cell_pair  => domain%cell_pair

    angle      => enefunc%num_angle
    alist      => enefunc%angle_list
    force      => enefunc%angle_force_const
    theta      => enefunc%angle_theta_min
    ubforce    => enefunc%urey_force_const
    ubrmin     => enefunc%urey_rmin

    nangl_lst  = domain_index%nangl_list
    nangl_p    = par%num_angles


    do iangl = 1, nangl_lst

      i = iangl_list(iangl)

      list(1) = cm_angl_list(1, i)
      list(2) = cm_angl_list(2, i)
      list(3) = cm_angl_list(3, i)
      ci1 = cm_atom_cls_name(list(1))
      ci2 = cm_atom_cls_name(list(2))
      ci3 = cm_atom_cls_name(list(3))
      ri1 = cm_residue_name(list(1))
      ri2 = cm_residue_name(list(2))
      ri3 = cm_residue_name(list(3))

      if (ri1(1:4) /= 'TIP3' .and. ri2(1:4) /= 'TIP3' .and. &
          ri3(1:4) /= 'TIP3') then

        icel1 = domain%id_g2l(1, list(1))
        icel2 = domain%id_g2l(1, list(3))

        if (icel1 /= 0 .and. icel2 /= 0) then

          icel_local = cell_pair(icel1, icel2)
       
          if (icel_local >= 1 .and. icel_local <= ncel) then

            do j = 1, nangl_p
              if ((ci1 == par%angl_atom_cls(1, j) .and. &
                   ci2 == par%angl_atom_cls(2, j) .and. &
                   ci3 == par%angl_atom_cls(3, j)) .or. &
                  (ci1 == par%angl_atom_cls(3, j) .and. &
                   ci2 == par%angl_atom_cls(2, j) .and. &
                   ci3 == par%angl_atom_cls(1, j))) then

                angle(icel_local) = angle(icel_local) + 1
                alist(1:3,angle(icel_local),icel_local) = list(1:3)
              force  (angle(icel_local),icel_local) = par%angl_force_const(j)
              theta  (angle(icel_local),icel_local) = par%angl_theta_min(j)*RAD
              ubforce(angle(icel_local),icel_local) = par%angl_ub_force_const(j)
              ubrmin (angle(icel_local),icel_local) = par%angl_ub_rmin(j)
                exit

              end if
            end do

            if (j == nangl_p + 1) &
              write(MsgOut,*) &
                'Setup_Enefunc_Angl_Constraint_Peer> not found ANGL: [', &
                ci1, ']-[', ci2, ']-[', ci3, '] in parameter file. (ERROR)'

          end if

        end if

      end if

    end do

    return

  end subroutine setup_enefunc_angl_constraint_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_restraints_peer
  !> @brief        peer routine of 
  !                         sp_enefunc_restraint_mod::setup_enefunc_restraint()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_restraints_peer(domain_index, domain, enefunc)

    ! formal arguments
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc


    !write(MsgOut,'(a)') '    Setup_Enefunc_Restraint_Peer_> '

    if (.not. enefunc%restraint) return


    call setup_enefunc_rest_domain_peer(domain_index, domain, enefunc)

    return

  end subroutine setup_enefunc_restraints_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_rest_domain_peer
  !> @brief        peer routine of 
  !                    sp_enefunc_restraints_mod::setup_enefunc_rest_domain()
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_rest_domain_peer(domain_index, domain, enefunc)

    ! formal arguments
    type(s_domain_index), target, intent(in)    :: domain_index
    type(s_domain),       target, intent(in)    :: domain
    type(s_enefunc),      target, intent(inout) :: enefunc

    ! local variable
    integer                       :: i, k, ig, iatm, icel
    integer                       :: iselatoms, nselatoms_lst

    real(wp),             pointer :: const(:,:)
    real(wp),             pointer :: restraint_force(:,:,:)
    real(wp),             pointer :: restraint_coord(:,:,:)
    integer,              pointer :: id_g2l(:,:)
    integer,              pointer :: ncel, num_funcs
    integer,              pointer :: grouplist(:,:), bondslist(:,:)
    integer,              pointer :: nrestraint(:), restraint_atom(:,:)
    integer,              pointer :: iselatoms_lst(:)


    !write(MsgOut,'(a)') '    Setup_Enefunc_Rest_Domain_Peer_> '

    ncel            => domain%num_cell_local
    id_g2l          => domain%id_g2l

    num_funcs       => enefunc%num_restraintfuncs
    grouplist       => enefunc%restraint_grouplist
    bondslist       => enefunc%restraint_bondslist
    const           => enefunc%restraint_const
    nrestraint      => enefunc%num_restraint

    ! check the size for distance restraint
    !
    k = 0
    do i = 1, num_funcs

      if (enefunc%restraint_kind(i) == RestraintsFuncDIST     .or. &
          enefunc%restraint_kind(i) == RestraintsFuncDISTCOM  .or. &
          enefunc%restraint_kind(i) == RestraintsFuncANGLE    .or. &
          enefunc%restraint_kind(i) == RestraintsFuncANGLECOM .or. &
          enefunc%restraint_kind(i) == RestraintsFuncDIHED    .or. &
          enefunc%restraint_kind(i) == RestraintsFuncDIHEDCOM ) then

        do ig = 1, enefunc%restraint_funcgrp(i)

          iselatoms_lst => domain_index%selatoms_list(grouplist(ig,i))%i
          nselatoms_lst =  domain_index%selatoms_list(grouplist(ig,i))%n

          k = k + nselatoms_lst

        end do

      end if

    end do
    enefunc%num_atoms_bonds_restraint = k


    call alloc_enefunc(enefunc, EneFuncRest, ncel, 1)

    restraint_atom  => enefunc%restraint_atom
    restraint_force => enefunc%restraint_force
    restraint_coord => enefunc%restraint_coord

    ! position restraint
    !
    do i = 1, num_funcs

      if (enefunc%restraint_kind(i) == RestraintsFuncPOSI) then

        iselatoms_lst => domain_index%selatoms_list(grouplist(1,i))%i
        nselatoms_lst =  domain_index%selatoms_list(grouplist(1,i))%n

        do iselatoms = 1, nselatoms_lst

          iatm = cm_selatoms(iselatoms_lst(iselatoms), grouplist(1,i))
          icel = id_g2l(1, iatm)

          if (icel > 0 .and. icel <= ncel) then
            nrestraint(icel) = nrestraint(icel) + 1
            restraint_atom (    nrestraint(icel),icel) = iatm
            restraint_force(1:4,nrestraint(icel),icel) = const(1:4,i)
            restraint_coord(1,nrestraint(icel),icel) = cm_atom_refcoord(1,iatm)
            restraint_coord(2,nrestraint(icel),icel) = cm_atom_refcoord(2,iatm)
            restraint_coord(3,nrestraint(icel),icel) = cm_atom_refcoord(3,iatm)
          end if

        end do

      end if

    end do


    ! distance restraint
    !

    ! ..skip

    return

  end subroutine setup_enefunc_rest_domain_peer

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    alloc_id_g2l
  !> @brief        allocate domain%id_g2l(:,:)
  !! @authors      NT
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine alloc_id_g2l(domain, natom)

    ! formal arguments
    type(s_domain),          intent(inout) :: domain
    integer,                 intent(in)    :: natom


    if (allocated(domain%id_g2l)) then
      if (size(domain%id_g2l(1,:)) /= natom) &
        deallocate(domain%id_g2l)
    end if

    if (.not. allocated(domain%id_g2l)) &
      allocate(domain%id_g2l(2, natom))

    domain%id_g2l(1:2,1:natom) = 0

    return

  end subroutine alloc_id_g2l

end module pr_setup_spdyn_peer_mod
