!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_dynvars_str_mod
!> @brief   structure of dynvars
!! @authors Yuji Sugita (YS), Takaharu Mori (TM), Chigusa Kobayashi (CK)
!! @note    dynvars include coord, force, energy, boxsize, pressure.
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_dynvars_str_mod

  use sp_energy_str_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_dynvars
    type(s_energy)                :: energy

    real(wp)                      :: virial(3,3)
    real(wp)                      :: virial_const(3,3)
    real(wp)                      :: virial_extern(3,3)

    integer                       :: step
    real(wp)                      :: time
    real(wp)                      :: total_pene
    real(wp)                      :: total_kene
    real(wp)                      :: total_energy
    real(wp)                      :: temperature
    real(wp)                      :: rms_gradient
    real(wp)                      :: hfc_kene
    real(wp)                      :: virial_kene

    real(wp)                      :: volume
    real(wp)                      :: internal_virial
    real(wp)                      :: internal_pressure
    real(wp)                      :: external_virial
    real(wp)                      :: external_pressure

    real(wp)                      :: thermostat_momentum
    real(wp)                      :: thermostat_momentum_ref
    real(wp)                      :: barostat_momentum(3)
    real(wp)                      :: barostat_momentum_ref(3)

    real(wp)                      :: box_size_x
    real(wp)                      :: box_size_y
    real(wp)                      :: box_size_z
    real(wp)                      :: pressure_xx
    real(wp)                      :: pressure_yy
    real(wp)                      :: pressure_zz
  end type s_dynvars

  ! subroutines
  public :: init_dynvars

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    init_dynvars
  !> @brief        initialize energy terms in dynvars
  !! @authors      YS, TM, CK
  !! @param[out]   dynvars : dynamic variable information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_dynvars(dynvars)

    ! formal arguments
    type(s_dynvars),         intent(inout) :: dynvars


    call init_energy (dynvars%energy)

    dynvars%virial(1:3,1:3)            = 0.0_wp
    dynvars%virial_const(1:3,1:3)      = 0.0_wp
    dynvars%virial_extern(1:3,1:3)     = 0.0_wp

    dynvars%step                       = 0
    dynvars%time                       = 0.0_wp
    dynvars%total_pene                 = 0.0_wp
    dynvars%total_kene                 = 0.0_wp
    dynvars%total_energy               = 0.0_wp
    dynvars%temperature                = 0.0_wp
    dynvars%rms_gradient               = 0.0_wp
    dynvars%hfc_kene                   = 0.0_wp
    dynvars%virial_kene                = 0.0_wp

    dynvars%volume                     = 0.0_wp
    dynvars%internal_virial            = 0.0_wp
    dynvars%internal_pressure          = 0.0_wp
    dynvars%external_virial            = 0.0_wp
    dynvars%external_pressure          = 0.0_wp

    dynvars%thermostat_momentum        = 0.0_wp
    dynvars%thermostat_momentum_ref    = 0.0_wp
    dynvars%barostat_momentum(1:3)     = 0.0_wp
    dynvars%barostat_momentum_ref(1:3) = 0.0_wp

    dynvars%box_size_x                 = 0.0_wp
    dynvars%box_size_y                 = 0.0_wp
    dynvars%box_size_z                 = 0.0_wp
    dynvars%pressure_xx                = 0.0_wp
    dynvars%pressure_yy                = 0.0_wp
    dynvars%pressure_zz                = 0.0_wp

    return

  end subroutine init_dynvars

end module sp_dynvars_str_mod
