!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_enefunc_str_mod
!> @brief   structure of energy functions
!! @authors Yuji Sugita (YS), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_enefunc_str_mod

  use sp_domain_str_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_table_enefunc

    ! General table
    logical                       :: table
    logical                       :: water_table
    integer                       :: atom_cls_no_O
    integer                       :: atom_cls_no_H
    integer                       :: table_order
    real(wp)                      :: density
    real(wp)                      :: charge_O
    real(wp)                      :: charge_H
    real(wp)                      :: mass_O
    real(wp)                      :: mass_H

    real(wp),         allocatable :: table_ene(:)
    real(wp),         allocatable :: table_grad(:)
    real(wp),         allocatable :: table_ecor(:)
    real(wp),         allocatable :: table_decor(:)
    real(wp),         allocatable :: dtable_ecor(:)
    real(wp),         allocatable :: dtable_decor(:)

    ! Water
    character(5)                  :: water_model
    integer                       :: num_water
    integer                       :: num_solute
    integer,          allocatable :: water_list(:,:)
    integer,          allocatable :: solute_list(:)
    integer,          allocatable :: solute_list_inv(:)

    real(wp),         allocatable :: table_ene_WW(:,:)
    real(wp),         allocatable :: table_de_WW(:,:)

  end type s_table_enefunc

  type, public :: s_enefunc

    integer                       :: forcefield
    integer                       :: output_style

    integer                       :: num_bond_all
    integer                       :: num_angl_all
    integer                       :: num_dihe_all
    integer                       :: num_impr_all
    integer                       :: num_cmap_all
    integer                       :: num_excl_all
    integer                       :: num_nb14_all

    integer                       :: num_atom_cls
    integer                       :: num_atoms_ref
    integer                       :: num_restraintgroups
    integer                       :: num_restraintfuncs
    integer                       :: max_restraint_numatoms
    integer                       :: max_restraint_numgrps

    ! base
    integer,          allocatable :: num_bond(:)
    integer,          allocatable :: num_angle(:)
    integer,          allocatable :: num_dihedral(:)
    integer,          allocatable :: num_improper(:)
    integer,          allocatable :: num_cmap(:)
    integer,          allocatable :: num_restraint(:)

    ! bond (size = num_local_bond for each cell)
    integer,          allocatable :: bond_list(:,:,:)
    real(wp),         allocatable :: bond_force_const(:,:)
    real(wp),         allocatable :: bond_dist_min(:,:)
    integer,          allocatable :: bond_kind(:,:)

    ! angle (size = num_local_angle for each cell)
    integer,          allocatable :: angle_list(:,:,:)
    real(wp),         allocatable :: angle_force_const(:,:)
    real(wp),         allocatable :: angle_theta_min(:,:)
    real(wp),         allocatable :: urey_force_const(:,:)
    real(wp),         allocatable :: urey_rmin(:,:)
    integer,          allocatable :: angle_kind(:,:)

    ! dihedral (size = num_local_dihedral for each cell)
    integer,          allocatable :: dihe_list(:,:,:)
    real(wp),         allocatable :: dihe_force_const(:,:)
    integer,          allocatable :: dihe_periodicity(:,:)
    real(wp),         allocatable :: dihe_phase(:,:)
    integer,          allocatable :: dihe_kind(:,:)

    ! improper (size = num_local_improper for each cell)
    integer,          allocatable :: impr_list(:,:,:)
    real(wp),         allocatable :: impr_force_const(:,:)
    integer,          allocatable :: impr_periodicity(:,:)
    real(wp),         allocatable :: impr_phase(:,:)

    ! cmap
    integer,          allocatable :: cmap_list(:,:,:)
    integer,          allocatable :: cmap_resolution(:)
    integer,          allocatable :: cmap_type(:,:)
    real(wp),         allocatable :: cmap_coef(:,:,:,:,:)

    ! non-bonded (size = num_atom_cls)
    integer,          allocatable :: nonb_atom_cls(:)
    real(wp),         allocatable :: nb14_lj6(:,:)
    real(wp),         allocatable :: nb14_lj12(:,:)
    real(wp),         allocatable :: nonb_lj6(:,:)
    real(wp),         allocatable :: nonb_lj12(:,:)

    ! non-bonded (size = num_atoms)
    integer,          allocatable :: num_nonb_excl(:,:)
    integer,          allocatable :: num_nonb_excl1(:,:) !RN: 1=> ???
    integer,          allocatable :: num_nb14_calc(:,:)
    integer,          allocatable :: num_nb14_calc1(:,:)
    integer,          allocatable :: nonb_list(:,:,:)
    integer,          allocatable :: nonb_list1(:,:,:)
    integer,          allocatable :: nb14_list(:,:,:)
    integer,          allocatable :: nb14_list1(:,:,:)
    integer,          allocatable :: num_excl_total(:)
    integer,          allocatable :: num_excl_total1(:)
    integer,          allocatable :: num_nb14_total(:)
    integer,          allocatable :: num_nb14_total1(:)

    ! non-bonded list (size = num_atoms)
    integer,          allocatable :: nonb_excl_list(:,:)
    integer,          allocatable :: nonb_excl_list1(:,:)
    integer,          allocatable :: nb14_calc_list(:,:)
    integer,          allocatable :: nb14_calc_list1(:,:)

    real(wp)                      :: switchdist
    real(wp)                      :: cutoffdist
    real(wp)                      :: pairlistdist
    real(wp)                      :: dielec_const

    logical                       :: pme_use
    real(wp)                      :: pme_alpha
    integer                       :: pme_ngrid_x
    integer                       :: pme_ngrid_y
    integer                       :: pme_ngrid_z
    integer                       :: pme_nspline

    ! flag for position restraint 
    logical                       :: restraint_posi
    ! restraint group (size = num_restraintgroups)
    integer,          allocatable :: restraint_numatoms(:)
    integer,          allocatable :: restraint_atomlist(:,:)
    integer,          allocatable :: restraint_bondslist(:,:)
    real(wp),         allocatable :: restraint_masscoef(:,:)

    ! restraint func (size = num_restraintfuncs)
    integer,          allocatable :: restraint_kind(:)
    integer,          allocatable :: restraint_grouplist(:,:)
    integer,          allocatable :: restraint_funcgrp(:)
    integer,          allocatable :: restraint_exponent_func(:)
    integer,          allocatable :: restraint_exponent_dist(:,:)
    real(wp),         allocatable :: restraint_weight_dist(:,:)
    real(wp),         allocatable :: restraint_const(:,:)
    real(wp),         allocatable :: restraint_ref(:,:)
    real(wp),         allocatable :: restraint_wcom1(:,:)
    real(wp),         allocatable :: restraint_wcom2(:,:)
    real(wp),         allocatable :: restraint_wcom3(:,:)
    real(wp),         allocatable :: restraint_wcom4(:,:)
    real(wp),         allocatable :: restraint_wcom5(:,:)
    real(wp),         allocatable :: restraint_wtmp(:,:)
    real(wp),         allocatable :: restraint_wdrt(:)

    ! restraint func (size = num_atoms_ref)
    real(wp),         allocatable :: restraint_refcoord(:,:)

    ! restraint
    logical                       :: restraint
    logical                       :: local_restraint
    integer                       :: num_atoms_bonds_restraint
    integer,          allocatable :: restraint_atom(:,:)
    integer,          allocatable :: restraint_bondslist_to_atomlist(:)
    real(wp),         allocatable :: restraint_force(:,:,:)
    real(wp),         allocatable :: restraint_coord(:,:,:)
    real(wp),         allocatable :: restraint_bonds_coord(:,:)
    real(wp),         allocatable :: restraint_bonds_force(:,:)

    ! update
    integer,          allocatable :: bond_exit(:)
    integer,          allocatable :: bond_add(:)
    integer,          allocatable :: bond_exit_index(:,:)
    integer,          allocatable :: buf_bond_integer(:,:,:)
    real(wp),         allocatable :: buf_bond_real(:,:,:)

    integer,          allocatable :: angle_exit(:)
    integer,          allocatable :: angle_add(:)
    integer,          allocatable :: angle_exit_index(:,:)
    integer,          allocatable :: buf_angle_integer(:,:,:)
    real(wp),         allocatable :: buf_angle_real(:,:,:)

    integer,          allocatable :: dihed_exit(:)
    integer,          allocatable :: dihed_add(:)
    integer,          allocatable :: dihed_exit_index(:,:)
    integer,          allocatable :: buf_dihed_integer(:,:,:)
    real(wp),         allocatable :: buf_dihed_real(:,:,:)

    integer,          allocatable :: impr_exit(:)
    integer,          allocatable :: impr_add(:)
    integer,          allocatable :: impr_exit_index(:,:)
    integer,          allocatable :: buf_impr_integer(:,:,:)
    real(wp),         allocatable :: buf_impr_real(:,:,:)

    integer,          allocatable :: cmap_exit(:)
    integer,          allocatable :: cmap_add(:)
    integer,          allocatable :: cmap_exit_index(:,:)
    integer,          allocatable :: buf_cmap_integer(:,:,:)

    integer,          allocatable :: restraint_exit(:)
    integer,          allocatable :: restraint_add(:)
    integer,          allocatable :: restraint_exit_index(:,:)
    integer,          allocatable :: buf_restraint_integer(:,:)
    real(wp),         allocatable :: buf_restraint_real(:,:,:)

    type(s_table_enefunc)         :: table

    logical                       :: force_switch

  end type s_enefunc

  ! parameter for allocatable variables
  integer,      public, parameter :: EneFuncBase         = 1
  integer,      public, parameter :: EneFuncBond         = 2
  integer,      public, parameter :: EneFuncAngl         = 3
  integer,      public, parameter :: EneFuncDihe         = 4
  integer,      public, parameter :: EneFuncImpr         = 5
  integer,      public, parameter :: EneFuncCmap         = 6
  integer,      public, parameter :: EneFuncNbon         = 7
  integer,      public, parameter :: EneFuncNonb         = 8
  integer,      public, parameter :: EneFuncNonbList     = 9
  integer,      public, parameter :: EneFuncRefg         = 10
  integer,      public, parameter :: EneFuncReff         = 11
  integer,      public, parameter :: EneFuncRefc         = 12
  integer,      public, parameter :: EneFuncRest         = 13
  integer,      public, parameter :: EneFuncBondCell     = 14
  integer,      public, parameter :: EneFuncRestCell     = 15
  integer,      public, parameter :: EneFuncTableWat     = 16
  integer,      public, parameter :: EneFuncTableSol     = 17
  integer,      public, parameter :: EneFuncTblWatDomain = 18
  integer,      public, parameter :: EneFuncTableDomain  = 19

  ! parameters (forcefield)
  integer,      public, parameter :: ForcefieldCHARMM    = 1
  character(*), public, parameter :: ForceFieldTypes(1)  = (/'CHARMM  '/)

  ! parameters (output style)
  integer,      public, parameter :: OutputStyleGENESIS  = 1
  integer,      public, parameter :: OutputStyleCHARMM   = 2
  integer,      public, parameter :: OutputStyleNAMD     = 3
  character(*), public, parameter :: OutputStyleTypes(3) = (/'GENESIS ', &
                                                             'CHARMM  ', &
                                                             'NAMD    '/)

  ! variables for maximum numbers in one cell
  integer,      public            :: MaxBond   = 200
  integer,      public            :: MaxAngle  = 350
  integer,      public            :: MaxDihe   = 700
  integer,      public            :: MaxImpr   = 200
  integer,      public            :: MaxCmap   = 20
  integer,      public            :: MaxRest   = 100
  integer,      public            :: MaxExcl   = 36

  integer,      public            :: BondMove  = 50
  integer,      public            :: AngleMove = 100
  integer,      public            :: DiheMove  = 300
  integer,      public            :: ImprMove  = 50
  integer,      public            :: CmapMove  = 10
  integer,      public            :: RestMove  = 50

  ! global variables
  integer,      public            :: table_maxpoint

  integer,              parameter :: MaxAtomCls = 1000
  real(wp),     public            :: lj_coef(2,MaxAtomCls)

  ! subroutines
  public  :: init_enefunc
  public  :: alloc_enefunc
  public  :: dealloc_enefunc
  public  :: dealloc_enefunc_all

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    init_enefunc
  !> @brief        initialize energy functions information
  !! @authors      YS, CK
  !! @param[out]   enefunc  : structure of potential energy function
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine init_enefunc(enefunc)

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc

    
    enefunc%forcefield             = ForcefieldCHARMM
    enefunc%output_style           = OutputStyleCHARMM

    enefunc%num_atom_cls           = 0
    enefunc%num_atoms_ref          = 0
    enefunc%num_restraintgroups    = 0
    enefunc%num_restraintfuncs     = 0
    enefunc%max_restraint_numatoms = 0

    enefunc%switchdist             = 0.0_wp
    enefunc%cutoffdist             = 0.0_wp
    enefunc%pairlistdist           = 0.0_wp
    enefunc%dielec_const           = 0.0_wp

    enefunc%pme_use                = .false.
    enefunc%pme_alpha              = 0.0_wp
    enefunc%pme_ngrid_x            = 0
    enefunc%pme_ngrid_y            = 0
    enefunc%pme_ngrid_z            = 0
    enefunc%pme_nspline            = 0

    enefunc%restraint_posi         = .false.
    
    enefunc%table%table            = .false.
    enefunc%table%water_table      = .false.
    enefunc%table%atom_cls_no_O    = 0
    enefunc%table%atom_cls_no_H    = 0
    enefunc%table%table_order      = 0
    enefunc%table%density          = 0.0_wp
    enefunc%table%charge_O         = 0.0_wp
    enefunc%table%charge_H         = 0.0_wp
    enefunc%table%mass_O           = 0.0_wp
    enefunc%table%mass_H           = 0.0_wp
    enefunc%table%water_model      = ''
    enefunc%table%num_water        = 0
    enefunc%table%num_solute       = 0

    return

  end subroutine init_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    alloc_enefunc
  !> @brief        allocate energy functions information
  !! @authors      YS, CK
  !! @param[out]   enefunc   : potential energy functions information
  !! @param[in]    variable  : selected variable
  !! @param[in]    var_size  : size of the selected variable
  !! @param[in]    var_size1 : 2nd size of the selected variable (optional)
  !! @param[in]    var_size2 : 3rd size of the selected variable (optional)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine alloc_enefunc(enefunc, variable, var_size, var_size1, var_size2)

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc
    integer,                 intent(in)    :: variable
    integer,                 intent(in)    :: var_size
    integer,       optional, intent(in)    :: var_size1
    integer,       optional, intent(in)    :: var_size2

    ! local variables
    integer                  :: alloc_stat
    integer                  :: dealloc_stat
    integer                  :: var_size3


    alloc_stat   = 0
    dealloc_stat = 0

    ! allocate selected variables
    !
    select case (variable)

    case (EneFuncBase)

      if (allocated(enefunc%num_bond)) then
        if (size(enefunc%num_bond(:)) /= var_size) &
          deallocate(enefunc%num_bond,     &
                     enefunc%num_angle,    &
                     enefunc%num_dihedral, &
                     enefunc%num_improper, &
                     enefunc%num_cmap,     &
                     enefunc%num_restraint)
      end if

      if (.not. allocated(enefunc%num_bond)) &
        allocate(enefunc%num_bond     (var_size), &
                 enefunc%num_angle    (var_size), &
                 enefunc%num_dihedral (var_size), &
                 enefunc%num_improper (var_size), &
                 enefunc%num_cmap     (var_size), &
                 enefunc%num_restraint(var_size), &
                 stat = alloc_stat)

      enefunc%num_bond     (1:var_size) = 0
      enefunc%num_angle    (1:var_size) = 0
      enefunc%num_dihedral (1:var_size) = 0
      enefunc%num_improper (1:var_size) = 0
      enefunc%num_cmap     (1:var_size) = 0
      enefunc%num_restraint(1:var_size) = 0

    case (EneFuncBond)

      if (allocated(enefunc%bond_list)) then
        if (size(enefunc%bond_list(1,1,:)) /= var_size) &
          deallocate(enefunc%bond_list,        &
                     enefunc%bond_force_const, &
                     enefunc%bond_dist_min,    &
                     enefunc%bond_kind,        &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%bond_list)) &
        allocate(enefunc%bond_list       (2, MaxBond, var_size), &
                 enefunc%bond_force_const(MaxBond, var_size),    &
                 enefunc%bond_dist_min   (MaxBond, var_size),    &
                 enefunc%bond_kind       (MaxBond, var_size),    &
                 stat = alloc_stat)

      enefunc%bond_list       (1:2, 1:MaxBond, 1:var_size) = 0
      enefunc%bond_force_const(1:MaxBond, 1:var_size)      = 0.0_wp
      enefunc%bond_dist_min   (1:MaxBond, 1:var_size)      = 0.0_wp
      enefunc%bond_kind       (1:MaxBond, 1:var_size)      = 0

    case (EneFuncAngl)

      if (allocated(enefunc%angle_list)) then
        if (size(enefunc%angle_list(1,1,:)) /= var_size) &
          deallocate(enefunc%angle_list,        &
                     enefunc%angle_force_const, &
                     enefunc%angle_theta_min,   &
                     enefunc%urey_force_const,  &
                     enefunc%urey_rmin,         &
                     enefunc%angle_kind,        &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%angle_list)) &
        allocate(enefunc%angle_list       (3, MaxAngle, var_size), &
                 enefunc%angle_force_const(MaxAngle, var_size),    &
                 enefunc%angle_theta_min  (MaxAngle, var_size),    &
                 enefunc%urey_force_const (MaxAngle, var_size),    &
                 enefunc%urey_rmin        (MaxAngle, var_size),    &
                 enefunc%angle_kind       (MaxAngle, var_size),    &
                 stat = alloc_stat)

      enefunc%angle_list       (1:3, 1:MaxAngle, 1:var_size) = 0
      enefunc%angle_force_const(1:MaxAngle, 1:var_size)      = 0.0_wp
      enefunc%angle_theta_min  (1:MaxAngle, 1:var_size)      = 0.0_wp
      enefunc%urey_force_const (1:MaxAngle, 1:var_size)      = 0.0_wp
      enefunc%urey_rmin        (1:MaxAngle, 1:var_size)      = 0.0_wp
      enefunc%angle_kind       (1:MaxAngle, 1:var_size)      = 0

    case (EneFuncDihe)

      if (allocated(enefunc%dihe_list)) then
        if (size(enefunc%dihe_list(1,1,:)) /= var_size) &
          deallocate(enefunc%dihe_list,        &
                     enefunc%dihe_force_const, &
                     enefunc%dihe_periodicity, &
                     enefunc%dihe_phase,       &
                     enefunc%dihe_kind,        &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%dihe_list)) &
        allocate(enefunc%dihe_list       (4, MaxDihe, var_size), &
                 enefunc%dihe_force_const(MaxDihe, var_size),    &
                 enefunc%dihe_periodicity(MaxDihe, var_size),    &
                 enefunc%dihe_phase      (MaxDihe, var_size),    &
                 enefunc%dihe_kind       (MaxDihe, var_size),    &
                 stat = alloc_stat)

      enefunc%dihe_list       (1:4, 1:MaxDihe, 1:var_size) = 0
      enefunc%dihe_force_const(1:MaxDihe, 1:var_size)      = 0.0_wp
      enefunc%dihe_periodicity(1:MaxDihe, 1:var_size)      = 0.0_wp
      enefunc%dihe_phase      (1:MaxDihe, 1:var_size)      = 0.0_wp
      enefunc%dihe_kind       (1:MaxDihe, 1:var_size)      = 0

    case (EneFuncImpr)

      if (allocated(enefunc%impr_list)) then
        if (size(enefunc%impr_list(1,1,:)) /= var_size) &
          deallocate(enefunc%impr_list,        &
                     enefunc%impr_force_const, &
                     enefunc%impr_periodicity, &
                     enefunc%impr_phase,       &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%impr_list)) &
        allocate(enefunc%impr_list       (4, MaxImpr, var_size), &
                 enefunc%impr_force_const(MaxImpr, var_size),    &
                 enefunc%impr_periodicity(MaxImpr, var_size),    &
                 enefunc%impr_phase      (MaxImpr, var_size),    &
                 stat = alloc_stat)

      enefunc%impr_list       (1:4, 1:MaxImpr, 1:var_size) = 0
      enefunc%impr_force_const(1:MaxImpr, 1:var_size)      = 0.0_wp
      enefunc%impr_periodicity(1:MaxImpr, 1:var_size)      = 0.0_wp
      enefunc%impr_phase      (1:MaxImpr, 1:var_size)      = 0.0_wp

    case (EneFuncCmap)

      if (allocated(enefunc%cmap_resolution)) then
        if (size(enefunc%cmap_list(1,1,:))   /= var_size .or. &
            size(enefunc%cmap_resolution(:)) /= var_size2)    &
          deallocate(enefunc%cmap_resolution, &
                     enefunc%cmap_list,       &
                     enefunc%cmap_type,       &
                     enefunc%cmap_coef,       &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%cmap_resolution)) &
        allocate(enefunc%cmap_resolution(var_size2),     &
                 enefunc%cmap_list(8,MaxCmap, var_size), &
                 enefunc%cmap_type(MaxCmap, var_size),   &
                 enefunc%cmap_coef(4, 4, var_size1, var_size1, var_size2), &
                 stat = alloc_stat)

      enefunc%cmap_resolution(1:var_size2)          = 0
      enefunc%cmap_list(1:8, 1:MaxCmap, 1:var_size) = 0
      enefunc%cmap_type(1:MaxCmap, 1:var_size)      = 0
      enefunc%cmap_coef(1:4, 1:4, 1:var_size1, 1:var_size1, 1:var_size2) = 0.0_wp

    case (EneFuncNbon)

      if (allocated(enefunc%nonb_atom_cls)) then
        if (size(enefunc%nonb_atom_cls(:)) /= var_size) &
          deallocate(enefunc%nonb_atom_cls, &
                     enefunc%nb14_lj6,      &
                     enefunc%nb14_lj12,     &
                     enefunc%nonb_lj6,      &
                     enefunc%nonb_lj12,     &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%nonb_atom_cls)) &
        allocate(enefunc%nonb_atom_cls(var_size),           &
                 enefunc%nb14_lj6     (var_size, var_size), &
                 enefunc%nb14_lj12    (var_size, var_size), &
                 enefunc%nonb_lj6     (var_size, var_size), &
                 enefunc%nonb_lj12    (var_size, var_size), &
                 stat = alloc_stat)

      enefunc%nonb_atom_cls(1:var_size)             = 0
      enefunc%nb14_lj6     (1:var_size, 1:var_size) = 0.0_wp
      enefunc%nb14_lj12    (1:var_size, 1:var_size) = 0.0_wp
      enefunc%nonb_lj6     (1:var_size, 1:var_size) = 0.0_wp
      enefunc%nonb_lj12    (1:var_size, 1:var_size) = 0.0_wp

    case (EneFuncNonb)

      if (allocated(enefunc%num_nonb_excl)) then
        if (size(enefunc%num_nonb_excl(1,:)) /= var_size1) &
          deallocate(enefunc%num_nonb_excl,   &
                     enefunc%num_nonb_excl1,  &
                     enefunc%num_nb14_calc,   &
                     enefunc%num_nb14_calc1,  &
                     enefunc%nonb_list,       &
                     enefunc%nonb_list1,      &
                     enefunc%nb14_list,       &
                     enefunc%nb14_list1,      &
                     enefunc%num_excl_total,  &
                     enefunc%num_excl_total1, &
                     enefunc%num_nb14_total,  &
                     enefunc%num_nb14_total1, &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%num_nonb_excl)) &
        allocate(enefunc%num_nonb_excl  (MaxAtom, var_size1),          &
                 enefunc%num_nonb_excl1 (MaxAtom, var_size),           &
                 enefunc%num_nb14_calc  (MaxAtom, var_size1),          &
                 enefunc%num_nb14_calc1 (MaxAtom, var_size),           &
                 enefunc%nonb_list      (MaxExcl, MaxAtom, var_size1), &
                 enefunc%nonb_list1     (MaxExcl, MaxAtom, var_size),  &
                 enefunc%nb14_list      (MaxExcl, MaxAtom, var_size1), &
                 enefunc%nb14_list1     (MaxExcl, MaxAtom, var_size),  &
                 enefunc%num_excl_total (var_size1),                   &
                 enefunc%num_excl_total1(var_size),                    &
                 enefunc%num_nb14_total (var_size1),                   &
                 enefunc%num_nb14_total1(var_size),                    &
                 stat = alloc_stat)

      enefunc%num_nonb_excl  (1:MaxAtom, 1:var_size1)             = 0
      enefunc%num_nonb_excl1 (1:MaxAtom, 1:var_size)              = 0
      enefunc%num_nb14_calc  (1:MaxAtom, 1:var_size1)             = 0
      enefunc%num_nb14_calc1 (1:MaxAtom, 1:var_size)              = 0
      enefunc%num_excl_total (1:var_size1)                        = 0
      enefunc%num_excl_total1(1:var_size)                         = 0
      enefunc%num_nb14_total (1:var_size1)                        = 0
      enefunc%num_nb14_total1(1:var_size)                         = 0

    case (EneFuncNonbList)

      if (allocated(enefunc%nonb_excl_list)) then
        if (size(enefunc%nonb_excl_list(1,:)) /= var_size1) &
          deallocate(enefunc%nonb_excl_list,  &
                     enefunc%nb14_calc_list,  &
                     enefunc%nonb_excl_list1, &
                     enefunc%nb14_calc_list1, &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%nonb_excl_list)) &
        allocate(enefunc%nonb_excl_list (MaxBond+MaxAngle, var_size1), &
                 enefunc%nb14_calc_list (MaxDihe, var_size1),          &
                 enefunc%nonb_excl_list1(MaxBond+MaxAngle, var_size),  &
                 enefunc%nb14_calc_list1(MaxDihe, var_size),           &
                 stat = alloc_stat)

      enefunc%nonb_excl_list (1:MaxBond+MaxAngle, 1:var_size1) = 0
      enefunc%nb14_calc_list (1:MaxDihe, 1:var_size1)          = 0
      enefunc%nonb_excl_list1(1:MaxBond+MaxAngle, 1:var_size)  = 0
      enefunc%nb14_calc_list1(1:MaxDihe, 1:var_size)           = 0

    case (EneFuncRefg)

      if (allocated(enefunc%restraint_numatoms)) then
        if (size(enefunc%restraint_numatoms) /= var_size) &
          deallocate(enefunc%restraint_numatoms,  &
                     enefunc%restraint_atomlist,  &
                     enefunc%restraint_bondslist, &
                     enefunc%restraint_masscoef,  &
                     enefunc%restraint_wcom3,     &
                     enefunc%restraint_wcom4,     &
                     enefunc%restraint_wcom5,     &
                     enefunc%restraint_wtmp,      &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%restraint_numatoms)) &
        allocate(enefunc%restraint_numatoms (var_size),                &
                 enefunc%restraint_atomlist (1:var_size1, 1:var_size), &
                 enefunc%restraint_bondslist(1:var_size1, 1:var_size), &
                 enefunc%restraint_masscoef (1:var_size1, 1:var_size), &
                 enefunc%restraint_wcom3    (1:3, 1:var_size1),        &
                 enefunc%restraint_wcom4    (1:3, 1:var_size1),        &
                 enefunc%restraint_wcom5    (1:3, 1:var_size1),        &
                 enefunc%restraint_wtmp     (1:var_size1, 1:var_size), &
                 stat = alloc_stat)

      enefunc%restraint_numatoms (1:var_size)              = 0
      enefunc%restraint_atomlist (1:var_size1, 1:var_size) = 0
      enefunc%restraint_bondslist(1:var_size1, 1:var_size) = 0
      enefunc%restraint_masscoef (1:var_size1, 1:var_size) = 0.0_wp
      enefunc%restraint_wcom3    (1:3, 1:var_size1)        = 0.0_wp
      enefunc%restraint_wcom4    (1:3, 1:var_size1)        = 0.0_wp
      enefunc%restraint_wcom5    (1:3, 1:var_size1)        = 0.0_wp
      enefunc%restraint_wtmp     (1:var_size1, 1:var_size) = 0.0_wp

    case (EneFuncReff)

      if (allocated(enefunc%restraint_kind)) then
        if (size(enefunc%restraint_kind) /= var_size) &
          deallocate(enefunc%restraint_kind,          &
                     enefunc%restraint_grouplist,     &
                     enefunc%restraint_const,         &
                     enefunc%restraint_ref,           &
                     enefunc%restraint_funcgrp,       &
                     enefunc%restraint_exponent_func, &
                     enefunc%restraint_exponent_dist, &
                     enefunc%restraint_weight_dist,   &
                     enefunc%restraint_wcom1,         &
                     enefunc%restraint_wcom2,         &
                     enefunc%restraint_wdrt,          &
                     stat = dealloc_stat)
      end if

      var_size3 = max(int(var_size1/2),1)

      if (.not. allocated(enefunc%restraint_kind)) &
        allocate(enefunc%restraint_kind         (var_size),                &
                 enefunc%restraint_grouplist    (1:var_size1, 1:var_size), &
                 enefunc%restraint_const        (1:4, 1:var_size),         &
                 enefunc%restraint_ref          (1:2, 1:var_size),         &
                 enefunc%restraint_funcgrp      (1:var_size),              &
                 enefunc%restraint_exponent_func(1:var_size),              &
                 enefunc%restraint_exponent_dist(1:var_size3, 1:var_size), &
                 enefunc%restraint_weight_dist  (1:var_size3, 1:var_size), &
                 enefunc%restraint_wcom1        (1:3, 1:var_size1),        &
                 enefunc%restraint_wcom2        (1:3, 1:var_size1),        &
                 enefunc%restraint_wdrt         (1:var_size3),             &
                 stat = alloc_stat)

      enefunc%restraint_kind         (1:var_size)              = 0
      enefunc%restraint_grouplist    (1:var_size1, 1:var_size) = 0
      enefunc%restraint_const        (1:4, 1:var_size)         = 0.0_wp
      enefunc%restraint_ref          (1:2, 1:var_size)         = 0.0_wp
      enefunc%restraint_funcgrp      (1:var_size)              = 0
      enefunc%restraint_exponent_func(1:var_size)              = 0
      enefunc%restraint_exponent_dist(1:var_size3, 1:var_size) = 0
      enefunc%restraint_weight_dist  (1:var_size3, 1:var_size) = 0.0_wp
      enefunc%restraint_wcom1        (1:3, 1:var_size1)        = 0.0_wp
      enefunc%restraint_wcom2        (1:3, 1:var_size1)        = 0.0_wp
      enefunc%restraint_wdrt         (1:var_size3)             = 0.0_wp

    case (EneFuncRefc)

      if (allocated(enefunc%restraint_refcoord)) then
        if (size(enefunc%restraint_refcoord) /= var_size) &
          deallocate(enefunc%restraint_refcoord, stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%restraint_refcoord)) &
        allocate(enefunc%restraint_refcoord(1:3, var_size), stat = alloc_stat)

      enefunc%restraint_refcoord(1:3, 1:var_size) = 0.0_wp

    case (EneFuncRest)

      if (allocated(enefunc%restraint_atom)) then
        if (size(enefunc%restraint_atom(1,:)) /= var_size) &
          deallocate(enefunc%restraint_atom,                  &
                     enefunc%restraint_force,                 &
                     enefunc%restraint_coord,                 &
                     enefunc%restraint_bondslist_to_atomlist, &
                     enefunc%restraint_bonds_coord,           &
                     enefunc%restraint_bonds_force,           &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%restraint_atom)) &
        allocate(enefunc%restraint_atom (MaxRest, var_size),         &
                 enefunc%restraint_force(4, MaxRest, var_size),      &
                 enefunc%restraint_coord(3, MaxRest, var_size),      &
                 enefunc%restraint_bondslist_to_atomlist(var_size1), &
                 enefunc%restraint_bonds_coord(3, var_size1),        &
                 enefunc%restraint_bonds_force(3, var_size1),        &
                 stat = alloc_stat)

      enefunc%restraint_atom (1:MaxRest, 1:var_size)       = 0
      enefunc%restraint_force(1:4, 1:MaxRest, 1:var_size)  = 0
      enefunc%restraint_coord(1:3, 1:MaxRest, 1:var_size)  = 0.0_wp
      enefunc%restraint_bondslist_to_atomlist(1:var_size1) = 0.0_wp
      enefunc%restraint_bonds_coord(1:3, 1:var_size1)      = 0.0_wp
      enefunc%restraint_bonds_force(1:3, 1:var_size1)      = 0.0_wp

    case (EneFuncBondCell)

      if (allocated(enefunc%bond_exit)) then
        if (size(enefunc%bond_exit(:)) /= var_size) &
          deallocate(enefunc%bond_exit,         &
                     enefunc%bond_add,          &
                     enefunc%bond_exit_index,   &
                     enefunc%buf_bond_integer,  &
                     enefunc%buf_bond_real,     &
                     enefunc%angle_exit,        &
                     enefunc%angle_add,         &
                     enefunc%angle_exit_index,  &
                     enefunc%buf_angle_integer, &
                     enefunc%buf_angle_real,    &
                     enefunc%dihed_exit,        &
                     enefunc%dihed_add,         &
                     enefunc%dihed_exit_index,  &
                     enefunc%buf_dihed_integer, &
                     enefunc%buf_dihed_real,    &
                     enefunc%impr_exit,         &
                     enefunc%impr_add,          &
                     enefunc%impr_exit_index,   &
                     enefunc%buf_impr_integer,  &
                     enefunc%buf_impr_real,     &
                     enefunc%cmap_exit,         &
                     enefunc%cmap_add,          &
                     enefunc%cmap_exit_index,   &
                     enefunc%buf_cmap_integer,  &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%bond_exit)) &
        allocate(enefunc%bond_exit        (var_size),                &
                 enefunc%bond_add         (var_size1),               &
                 enefunc%bond_exit_index  (BondMove, var_size),      &
                 enefunc%buf_bond_integer (3, BondMove, var_size1),  &
                 enefunc%buf_bond_real    (2, BondMove, var_size1),  &
                 enefunc%angle_exit       (var_size),                &
                 enefunc%angle_add        (var_size1),               &
                 enefunc%angle_exit_index (AngleMove, var_size),     &
                 enefunc%buf_angle_integer(4, AngleMove, var_size1), &
                 enefunc%buf_angle_real   (4, AngleMove, var_size1), &
                 enefunc%dihed_exit       (var_size),                &
                 enefunc%dihed_add        (var_size1),               &
                 enefunc%dihed_exit_index (DiheMove, var_size),      &
                 enefunc%buf_dihed_integer(6, DiheMove, var_size1),  &
                 enefunc%buf_dihed_real   (2, DiheMove, var_size1),  &
                 enefunc%impr_exit        (var_size),                &
                 enefunc%impr_add         (var_size1),               &
                 enefunc%impr_exit_index  (ImprMove, var_size),      &
                 enefunc%buf_impr_integer (4, ImprMove, var_size1),  &
                 enefunc%buf_impr_real    (2, ImprMove, var_size1),  &
                 enefunc%cmap_exit        (var_size),                &
                 enefunc%cmap_add         (var_size1),               &
                 enefunc%cmap_exit_index  (CmapMove, var_size),      &
                 enefunc%buf_cmap_integer (9, CmapMove, var_size1),  &
                 stat = alloc_stat)

      enefunc%bond_exit        (1:var_size)                    = 0
      enefunc%bond_add         (1:var_size1)                   = 0
      enefunc%bond_exit_index  (1:BondMove, 1:var_size)        = 0
      enefunc%buf_bond_integer (1:3, 1:BondMove, 1:var_size1)  = 0
      enefunc%buf_bond_real    (1:2, 1:BondMove, 1:var_size1)  = 0.0_wp
      enefunc%angle_exit       (1:var_size)                    = 0
      enefunc%angle_add        (1:var_size1)                   = 0
      enefunc%angle_exit_index (1:AngleMove, 1:var_size)       = 0
      enefunc%buf_angle_integer(1:4, 1:AngleMove, 1:var_size1) = 0
      enefunc%buf_angle_real   (1:4, 1:AngleMove, 1:var_size1) = 0.0_wp
      enefunc%dihed_exit       (1:var_size)                    = 0
      enefunc%dihed_add        (1:var_size1)                   = 0
      enefunc%dihed_exit_index (1:DiheMove, 1:var_size)        = 0
      enefunc%buf_dihed_integer(1:6, 1:DiheMove, 1:var_size1)  = 0
      enefunc%buf_dihed_real   (1:2, 1:DiheMove, 1:var_size1)  = 0.0_wp
      enefunc%impr_exit        (1:var_size)                    = 0
      enefunc%impr_add         (1:var_size1)                   = 0
      enefunc%impr_exit_index  (1:ImprMove, 1:var_size)        = 0
      enefunc%buf_impr_integer (1:4, 1:ImprMove, 1:var_size1)  = 0
      enefunc%buf_impr_real    (1:2, 1:ImprMove, 1:var_size1)  = 0.0_wp
      enefunc%cmap_exit        (1:var_size)                    = 0
      enefunc%cmap_add         (1:var_size1)                   = 0
      enefunc%cmap_exit_index  (1:CmapMove, 1:var_size)        = 0
      enefunc%buf_cmap_integer (1:9, 1:CmapMove, 1:var_size1)  = 0

    case (EneFuncRestCell)

      if (allocated(enefunc%restraint_exit)) then
        if (size(enefunc%restraint_exit(:)) /= var_size) &
          deallocate(enefunc%restraint_exit,        &
                     enefunc%restraint_add,         &
                     enefunc%restraint_exit_index,  &
                     enefunc%buf_restraint_integer, &
                     enefunc%buf_restraint_real,    &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%restraint_exit)) &
        allocate(enefunc%restraint_exit       (var_size),               &
                 enefunc%restraint_add        (var_size1),              &
                 enefunc%restraint_exit_index (RestMove, var_size),     &
                 enefunc%buf_restraint_integer(RestMove, var_size1),    &
                 enefunc%buf_restraint_real   (7, RestMove, var_size1), &
                 stat = alloc_stat)

      enefunc%restraint_exit       (1:var_size)                 = 0
      enefunc%restraint_add        (1:var_size1)                = 0
      enefunc%restraint_exit_index (1:RestMove, 1:var_size)     = 0
      enefunc%buf_restraint_integer(1:RestMove, 1:var_size1)    = 0
      enefunc%buf_restraint_real   (1:7, RestMove, 1:var_size1) = 0.0_wp

    case (EneFuncTableWat)

      if (allocated(enefunc%table%water_list)) then
         if (size(enefunc%table%water_list(1,:)) /= var_size) &
           deallocate(enefunc%table%water_list, stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%table%water_list)) &
        allocate(enefunc%table%water_list(3, var_size), stat = alloc_stat)

      enefunc%table%water_list(1:3, 1:var_size) = 0

    case (EneFuncTableSol)

      if (allocated(enefunc%table%solute_list)) then
         if (size(enefunc%table%solute_list(:)) /= var_size) &
           deallocate(enefunc%table%solute_list,      &
                      enefunc%table%solute_list_inv,  &
                      stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%table%solute_list)) &
        allocate(enefunc%table%solute_list    (var_size),  &
                 enefunc%table%solute_list_inv(var_size1), &
                 stat = alloc_stat)

      enefunc%table%solute_list    (1:var_size)  = 0
      enefunc%table%solute_list_inv(1:var_size1) = 0

    case (EneFuncTblWatDomain)

      if (allocated(enefunc%table%table_ene_WW)) then
        if (size(enefunc%table%table_ene(:)) /= 6*var_size) &
          deallocate(enefunc%table%table_ene,    &
                     enefunc%table%table_grad,   &
                     enefunc%table%table_ecor,   &
                     enefunc%table%table_decor,  &
                     enefunc%table%dtable_ecor,  &
                     enefunc%table%dtable_decor, &
                     enefunc%table%table_ene_WW, &
                     enefunc%table%table_de_WW,  &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%table%table_ene_WW)) &
        allocate(enefunc%table%table_ene   (6*var_size),    &
                 enefunc%table%table_grad  (6*var_size),    &
                 enefunc%table%table_ecor  (var_size),      &
                 enefunc%table%table_decor (var_size),      &
                 enefunc%table%dtable_ecor (var_size),      &
                 enefunc%table%dtable_decor(var_size),      &
                 enefunc%table%table_ene_WW(6*var_size, 3), &
                 enefunc%table%table_de_WW (2*var_size, 3), &
                 stat = alloc_stat)

      enefunc%table%table_ene   (1:6*var_size)      = 0.0_wp
      enefunc%table%table_grad  (1:6*var_size)      = 0.0_wp
      enefunc%table%table_ecor  (1:var_size)        = 0.0_wp
      enefunc%table%table_decor (1:var_size)        = 0.0_wp
      enefunc%table%dtable_ecor (1:var_size)        = 0.0_wp
      enefunc%table%dtable_decor(1:var_size)        = 0.0_wp
      enefunc%table%table_ene_WW(1:6*var_size, 1:3) = 0.0_wp
      enefunc%table%table_de_WW (1:2*var_size, 1:3) = 0.0_wp

    case (EneFuncTableDomain)

      if (allocated(enefunc%table%table_ene)) then
        if (size(enefunc%table%table_ene(:)) /= 6*var_size) &
          deallocate(enefunc%table%table_ene,    &
                     enefunc%table%table_grad,   &
                     enefunc%table%table_ecor,   &
                     enefunc%table%table_decor,  &
                     enefunc%table%dtable_ecor,  &
                     enefunc%table%dtable_decor, &
                     stat = dealloc_stat)
      end if

      if (.not. allocated(enefunc%table%table_ene)) &
        allocate(enefunc%table%table_ene   (6*var_size), &
                 enefunc%table%table_grad  (6*var_size), &
                 enefunc%table%table_ecor  (var_size),   &
                 enefunc%table%table_decor (var_size),   &
                 enefunc%table%dtable_ecor (var_size),   &
                 enefunc%table%dtable_decor(var_size),   &
                 stat = alloc_stat)

      enefunc%table%table_ene   (1:6*var_size) = 0.0_wp
      enefunc%table%table_grad  (1:6*var_size) = 0.0_wp
      enefunc%table%table_ecor  (1:var_size)   = 0.0_wp
      enefunc%table%table_decor (1:var_size)   = 0.0_wp
      enefunc%table%dtable_ecor (1:var_size)   = 0.0_wp
      enefunc%table%dtable_decor(1:var_size)   = 0.0_wp

    case default

      call error_msg('Alloc_Enefunc> bad variable')

    end select

    if (alloc_stat /= 0)   call error_msg_alloc
    if (dealloc_stat /= 0) call error_msg_dealloc    

    return

  end subroutine alloc_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_enefunc
  !> @brief        deallocate energy functions information
  !! @authors      YS, CK
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[in]    variable : selected variable
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_enefunc(enefunc, variable)

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc
    integer,                 intent(in)    :: variable

    ! local variables
    integer                  :: dealloc_stat

  
    dealloc_stat = 0

    select case (variable)

    case (EneFuncBase)

      if (allocated(enefunc%num_bond)) then
        deallocate(enefunc%num_bond,      &
                   enefunc%num_angle,     &
                   enefunc%num_dihedral,  &
                   enefunc%num_improper,  &
                   enefunc%num_cmap,      &
                   enefunc%num_restraint, &
                   stat = dealloc_stat)
      end if

    case (EneFuncBond)

      if (allocated(enefunc%bond_list)) then
        deallocate(enefunc%bond_list,        &
                   enefunc%bond_force_const, &
                   enefunc%bond_dist_min,    &
                   enefunc%bond_kind,        &
                   stat = dealloc_stat)
      end if

    case (EneFuncAngl)

      if (allocated(enefunc%angle_list)) then
        deallocate(enefunc%angle_list,        &
                   enefunc%angle_force_const, &
                   enefunc%angle_theta_min,   &
                   enefunc%urey_force_const,  &
                   enefunc%urey_rmin,         &
                   enefunc%angle_kind,        &
                   stat = dealloc_stat)
      end if

    case (EneFuncDihe)

      if (allocated(enefunc%dihe_list)) then
        deallocate(enefunc%dihe_list,        &
                   enefunc%dihe_force_const, &
                   enefunc%dihe_periodicity, &
                   enefunc%dihe_phase,       &
                   enefunc%dihe_kind,        &
                   stat = dealloc_stat)
      end if

    case (EneFuncImpr)

      if (allocated(enefunc%impr_list)) then
        deallocate(enefunc%impr_list,        &
                   enefunc%impr_force_const, &
                   enefunc%impr_periodicity, &
                   enefunc%impr_phase,       &
                   stat = dealloc_stat)
      end if

    case (EneFuncCmap)

      if (allocated(enefunc%cmap_resolution)) then
        deallocate(enefunc%cmap_resolution, &
                   enefunc%cmap_list, &
                   enefunc%cmap_type, &
                   enefunc%cmap_coef, &
                   stat = dealloc_stat)
      end if

    case (EneFuncNbon)

      if (allocated(enefunc%nonb_atom_cls)) then
        deallocate(enefunc%nonb_atom_cls, &
                   enefunc%nb14_lj6,      &
                   enefunc%nb14_lj12,     &
                   enefunc%nonb_lj6,      &
                   enefunc%nonb_lj12,     &
                   stat = dealloc_stat)
      end if

    case (EneFuncNonb)

      if (allocated(enefunc%num_nonb_excl)) then
        deallocate(enefunc%num_nonb_excl,   &
                   enefunc%num_nonb_excl1,  &
                   enefunc%num_nb14_calc,   &
                   enefunc%num_nb14_calc1,  &
                   enefunc%nonb_list,       &
                   enefunc%nonb_list1,      &
                   enefunc%nb14_list,       &
                   enefunc%nb14_list1,      &
                   enefunc%num_excl_total,  &
                   enefunc%num_excl_total1, &
                   enefunc%num_nb14_total,  &
                   enefunc%num_nb14_total1, &
                   stat = dealloc_stat)
      end if

    case (EneFuncNonbList)

      if (allocated(enefunc%nonb_excl_list)) then
        deallocate(enefunc%nonb_excl_list,  &
                   enefunc%nb14_calc_list,  &
                   enefunc%nonb_excl_list1, &
                   enefunc%nb14_calc_list1, &
                   stat = dealloc_stat)
      end if

    case (EneFuncRefg)

      if (allocated(enefunc%restraint_numatoms)) then
        deallocate(enefunc%restraint_numatoms, &
                   enefunc%restraint_atomlist, &
                   enefunc%restraint_masscoef, &
                   enefunc%restraint_wcom3,    &
                   enefunc%restraint_wcom4,    &
                   enefunc%restraint_wcom5,    &
                   enefunc%restraint_wtmp,     &
                   stat = dealloc_stat)
      end if

    case (EneFuncReff)

      if (allocated(enefunc%restraint_kind)) then
        deallocate(enefunc%restraint_kind,           &
                   enefunc%restraint_grouplist,      &
                   enefunc%restraint_const,          &
                   enefunc%restraint_ref,            &
                   enefunc%restraint_funcgrp,        &
                   enefunc%restraint_exponent_func,  &
                   enefunc%restraint_exponent_dist,  &
                   enefunc%restraint_weight_dist,    &
                   enefunc%restraint_wcom1,          &
                   enefunc%restraint_wcom2,          &
                   enefunc%restraint_wdrt,           &
                   stat = dealloc_stat)
       end if

    case (EneFuncRefc)

      if (allocated(enefunc%restraint_refcoord)) then
        deallocate(enefunc%restraint_refcoord, &
                   stat = dealloc_stat)
       end if

    case (EneFuncRest)

      if (allocated(enefunc%restraint_atom)) then
        deallocate(enefunc%restraint_atom,                  &
                   enefunc%restraint_force,                 &
                   enefunc%restraint_coord,                 &
                   enefunc%restraint_bondslist_to_atomlist, &
                   enefunc%restraint_bonds_coord,           &
                   enefunc%restraint_bonds_force,           &
                   stat = dealloc_stat)
      end if

    case (EneFuncBondCell)

      if (allocated(enefunc%bond_exit)) then
        deallocate(enefunc%bond_exit,         &
                   enefunc%bond_add,          &
                   enefunc%bond_exit_index,   &
                   enefunc%buf_bond_integer,  &
                   enefunc%buf_bond_real,     &
                   enefunc%angle_exit,        &
                   enefunc%angle_add,         &
                   enefunc%angle_exit_index,  &
                   enefunc%buf_angle_integer, &
                   enefunc%buf_angle_real,    &
                   enefunc%dihed_exit,        &
                   enefunc%dihed_add,         &
                   enefunc%dihed_exit_index,  &
                   enefunc%buf_dihed_integer, &
                   enefunc%buf_dihed_real,    &
                   enefunc%impr_exit,         &
                   enefunc%impr_add,          &
                   enefunc%impr_exit_index,   &
                   enefunc%buf_impr_integer,  &
                   enefunc%buf_impr_real,     &
                   enefunc%cmap_exit,         &
                   enefunc%cmap_add,          &
                   enefunc%cmap_exit_index,   &
                   enefunc%buf_cmap_integer,  &
                   stat = dealloc_stat)
      end if

    case (EneFuncRestCell)

      if (allocated(enefunc%restraint_exit)) then
        deallocate(enefunc%restraint_exit,        &
                   enefunc%restraint_add,         &
                   enefunc%restraint_exit_index,  &
                   enefunc%buf_restraint_integer, &
                   enefunc%buf_restraint_real,    &
                   stat = dealloc_stat)

      end if

    case (EneFuncTableWat)

      if (allocated(enefunc%table%water_list)) then
         deallocate(enefunc%table%water_list, &
                    stat = dealloc_stat)
      end if

    case (EneFuncTableSol)

      if (allocated(enefunc%table%solute_list)) then
         deallocate(enefunc%table%solute_list,     &
                    enefunc%table%solute_list_inv, &
                    stat = dealloc_stat)
      end if
    case (EneFuncTblWatDomain)

      if (allocated(enefunc%table%table_ene_WW)) then
        deallocate(enefunc%table%table_ene,    &
                   enefunc%table%table_grad,   &
                   enefunc%table%table_ecor,   &
                   enefunc%table%table_decor,  &
                   enefunc%table%table_ene_WW, &
                   enefunc%table%table_de_WW,  &
                   stat = dealloc_stat)
       end if

    case (EneFuncTableDomain)

      if (allocated(enefunc%table%table_ene)) then
        deallocate(enefunc%table%table_ene,   &
                   enefunc%table%table_grad,  &
                   enefunc%table%table_ecor,  &
                   enefunc%table%table_decor, &
                   stat = dealloc_stat)
       end if

    case default

      call error_msg('Alloc_Enefunc> bad variable')

    end select

    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_enefunc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_enefunc_all
  !> @brief        deallocate all energy functions information
  !! @authors      YS, CK
  !! @param[out]   enefunc : structure of potential energy function
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_enefunc_all(enefunc)

    ! formal arguments
    type(s_enefunc),         intent(inout) :: enefunc


    call dealloc_enefunc(enefunc, EneFuncBase)
    call dealloc_enefunc(enefunc, EneFuncBond)
    call dealloc_enefunc(enefunc, EneFuncAngl)
    call dealloc_enefunc(enefunc, EneFuncDihe)
    call dealloc_enefunc(enefunc, EneFuncImpr)
    call dealloc_enefunc(enefunc, EneFuncCmap)
    call dealloc_enefunc(enefunc, EneFuncNbon)
    call dealloc_enefunc(enefunc, EneFuncNonb)
    call dealloc_enefunc(enefunc, EneFuncNonbList)
    call dealloc_enefunc(enefunc, EneFuncRefg)
    call dealloc_enefunc(enefunc, EneFuncReff)
    call dealloc_enefunc(enefunc, EneFuncRefc)
    call dealloc_enefunc(enefunc, EneFuncRest)
    call dealloc_enefunc(enefunc, EneFuncBondCell)
    call dealloc_enefunc(enefunc, EneFuncRestCell)
    call dealloc_enefunc(enefunc, EneFuncTableWat)
    call dealloc_enefunc(enefunc, EneFuncTableSol)
    call dealloc_enefunc(enefunc, EneFuncTblWatDomain)
    call dealloc_enefunc(enefunc, EneFuncTableDomain)

    return

  end subroutine dealloc_enefunc_all

end module sp_enefunc_str_mod
